# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read the configuration file."""

__all__ = [
    'Configuration',
    'DISABLED',
    'config',
    ]


import os
import atexit

from configparser import ConfigParser
from contextlib import ExitStack
from pkg_resources import resource_filename
from systemimage.helpers import (
    Bag, as_loglevel, as_object, as_timedelta, makedirs, temporary_directory)


DISABLED = object()


def expand_path(path):
    return os.path.abspath(os.path.expanduser(path))


def port_value_converter(value):
    if value.lower() in ('disabled', 'disable'):
        return DISABLED
    result = int(value)
    if result < 0:
        raise ValueError(value)
    return result


class Configuration:
    def __init__(self):
        # Defaults.
        self.config_file = None
        self.service = Bag()
        self.system = Bag()
        ini_path = resource_filename('systemimage.data', 'client.ini')
        self.load(ini_path)
        self._override = False
        # 2013-10-14 BAW This is a placeholder for rendezvous between the
        # downloader and the D-Bus service.  When running udner D-Bus and we
        # get a `paused` signal from the download manager, we need this to
        # plumb through an UpdatePaused signal to our clients.  It rather
        # sucks that we need a global for this, but I can't get the plumbing
        # to work otherwise.  This seems like the least horrible place to
        # stash this global.
        self.dbus_service = None
        # Cache/overrides.
        self._device = None
        self._build_number = None
        self._channel = None
        self._tempdir = None
        self._resources = ExitStack()
        atexit.register(self._resources.close)

    def load(self, path, *, override=False):
        parser = ConfigParser()
        files_read = parser.read(path)
        if files_read != [path]:
            raise FileNotFoundError(path)
        self.config_file = path
        self.service.update(converters=dict(http_port=port_value_converter,
                                            https_port=port_value_converter,
                                            build_number=int),
                           **parser['service'])
        if (self.service.http_port is DISABLED and
            self.service.https_port is DISABLED):
            raise ValueError('Cannot disable both http and https ports')
        # Construct the HTTP and HTTPS base urls, which most applications will
        # actually use.  We do this in two steps, in order to support
        # disabling one or the other (but not both) protocols.
        if self.service.http_port == 80:
            http_base = 'http://{}'.format(self.service.base)
        elif self.service.http_port is DISABLED:
            http_base = None
        else:
            http_base = 'http://{}:{}'.format(
                self.service.base, self.service.http_port)
        # HTTPS.
        if self.service.https_port == 443:
            https_base = 'https://{}'.format(self.service.base)
        elif self.service.https_port is DISABLED:
            https_base = None
        else:
            https_base = 'https://{}:{}'.format(
                self.service.base, self.service.https_port)
        # Sanity check and final settings.
        if http_base is None:
            assert https_base is not None
            http_base = https_base
        if https_base is None:
            assert http_base is not None
            https_base = http_base
        self.service['http_base'] = http_base
        self.service['https_base'] = https_base
        try:
            self.system.update(converters=dict(timeout=as_timedelta,
                                               build_file=expand_path,
                                               loglevel=as_loglevel,
                                               settings_db=expand_path,
                                               tempdir=expand_path),
                              **parser['system'])
        except KeyError:
            # If we're overriding via a channel.ini file, it's okay if the
            # [system] section is missing.  However, the main configuration
            # ini file must include all sections.
            if not override:
                raise
        # Short-circuit, since we're loading a channel.ini file.
        self._override = override
        if override:
            return
        self.gpg = Bag(**parser['gpg'])
        self.updater = Bag(**parser['updater'])
        self.hooks = Bag(converters=dict(device=as_object,
                                         scorer=as_object,
                                         reboot=as_object),
                         **parser['hooks'])
        self.dbus = Bag(converters=dict(lifetime=as_timedelta),
                        **parser['dbus'])

    @property
    def build_number(self):
        if self._build_number is None:
            if self._override:
                return self.service.build_number
            else:
                try:
                    with open(self.system.build_file, encoding='utf-8') as fp:
                        return int(fp.read().strip())
                except FileNotFoundError:
                    return 0
        return self._build_number

    @build_number.setter
    def build_number(self, value):
        if not isinstance(value, int):
            raise ValueError(
                'integer is required, got: {}'.format(type(value).__name__))
        self._build_number = value

    @build_number.deleter
    def build_number(self):
        self._build_number = None

    @property
    def build_number_cli(self):
        return self._build_number

    @property
    def device(self):
        # It's safe to cache this.
        if self._device is None:
            self._device = self.hooks.device().get_device()
        return self._device

    @device.setter
    def device(self, value):
        self._device = value

    @property
    def channel(self):
        if self._channel is None:
            self._channel = self.service.channel
        return self._channel

    @channel.setter
    def channel(self, value):
        self._channel = value

    @property
    def tempdir(self):
        if self._tempdir is None:
            makedirs(self.system.tempdir)
            self._tempdir = self._resources.enter_context(
                temporary_directory(prefix='system-image-',
                                    dir=self.system.tempdir))
        return self._tempdir


# Define the global configuration object.  Normal use can be as simple as:
#
# from systemimage.config import config
# build_file = config.system.build_file
#
# In the test suite though, the actual configuration object can be easily
# patched by doing something like this:
#
# test_config = Configuration(...)
# with unittest.mock.patch('config._config', test_config):
#     run_test()
#
# and now every module which does the first code example will get build_file
# from the mocked Configuration instance.

_config = Configuration()

class _Proxy:
    def __getattribute__(self, name):
        return getattr(_config, name)
    def __setattr__(self, name, value):
        setattr(_config, name, value)
    def __delattr__(self, name):
        delattr(_config, name)


config = _Proxy()



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the DBus API mediator."""


__all__ = [
    'TestAPI',
    ]


import os
import unittest

from contextlib import ExitStack
from datetime import datetime, timedelta
from gi.repository import GLib
from systemimage.api import Mediator
from systemimage.config import Configuration, config
from systemimage.download import Canceled
from systemimage.gpg import SignatureError
from systemimage.testing.helpers import (
    ServerTestBase, chmod, configuration, copy, setup_index, sign,
    touch_build)
from unittest.mock import patch


MiB = 1024 * 1024


class TestAPI(ServerTestBase):
    INDEX_FILE = 'index_13.json'
    CHANNEL_FILE = 'channels_06.json'
    CHANNEL = 'stable'
    DEVICE = 'nexus7'

    @configuration
    def test_update_available(self):
        # Because our build number is lower than the latest available in the
        # index file, there is an update available.
        self._setup_server_keyrings()
        update = Mediator().check_for_update()
        self.assertTrue(update.is_available)

    @configuration
    def test_update_available_version(self):
        # An update is available.  What's the target version number?
        self._setup_server_keyrings()
        update = Mediator().check_for_update()
        self.assertEqual(update.version, '1600')

    @configuration
    def test_no_update_available_version(self):
        # No update is available, so the target version number is zero.
        self._setup_server_keyrings()
        touch_build(1600)
        update = Mediator().check_for_update()
        self.assertFalse(update.is_available)
        self.assertEqual(update.version, '')

    @configuration
    def test_no_update_available_at_latest(self):
        # Because our build number is equal to the latest available in the
        # index file, there is no update available.
        self._setup_server_keyrings()
        touch_build(1600)
        update = Mediator().check_for_update()
        self.assertFalse(update.is_available)

    @configuration
    def test_no_update_available_newer(self):
        # Because our build number is higher than the latest available in the
        # index file, there is no update available.
        self._setup_server_keyrings()
        touch_build(1700)
        update = Mediator().check_for_update()
        self.assertFalse(update.is_available)

    @configuration
    def test_get_details(self):
        # Get the details of an available update.
        self._setup_server_keyrings()
        # Index 14 has a more interesting upgrade path, and will yield a
        # richer description set.
        index_dir = os.path.join(self._serverdir, self.CHANNEL, self.DEVICE)
        index_path = os.path.join(index_dir, 'index.json')
        copy('index_14.json', index_dir, 'index.json')
        sign(index_path, 'device-signing.gpg')
        setup_index('index_14.json', self._serverdir, 'device-signing.gpg')
        # Get the descriptions.
        update = Mediator().check_for_update()
        self.assertTrue(update.is_available)
        self.assertEqual(update.size, 180009)
        self.assertEqual(len(update.descriptions), 3)
        # The first contains the descriptions for the full update.
        self.assertEqual(update.descriptions[0], {
            'description': 'Full B',
            'description-en': 'The full B',
            })
        # The first delta.
        self.assertEqual(update.descriptions[1], {
            'description': 'Delta B.1',
            'description-en_US': 'This is the delta B.1',
            'description-xx': 'XX This is the delta B.1',
            'description-yy': 'YY This is the delta B.1',
            'description-yy_ZZ': 'YY-ZZ This is the delta B.1',
            })
        # The second delta.
        self.assertEqual(update.descriptions[2], {
            'description': 'Delta B.2',
            'description-xx': 'Oh delta, my delta',
            'description-xx_CC': 'This hyar is the delta B.2',
            })

    @configuration
    def test_download(self):
        # After checking that an update is available, complete the update, but
        # don't reboot.
        self._setup_server_keyrings()
        mediator = Mediator()
        self.assertTrue(mediator.check_for_update())
        # Make sure a reboot did not get issued.
        got_reboot = False
        def reboot_mock(self):
            nonlocal got_reboot
            got_reboot = True
        with unittest.mock.patch(
                'systemimage.reboot.Reboot.reboot', reboot_mock):
            mediator.download()
        # No reboot got issued.
        self.assertFalse(got_reboot)
        # But the command file did get written, and all the files are present.
        path = os.path.join(config.updater.cache_partition, 'ubuntu_command')
        with open(path, 'r', encoding='utf-8') as fp:
            command = fp.read()
        self.assertMultiLineEqual(command, """\
load_keyring image-master.tar.xz image-master.tar.xz.asc
load_keyring image-signing.tar.xz image-signing.tar.xz.asc
load_keyring device-signing.tar.xz device-signing.tar.xz.asc
format system
mount system
update 6.txt 6.txt.asc
update 7.txt 7.txt.asc
update 5.txt 5.txt.asc
unmount system
""")
        self.assertEqual(set(os.listdir(config.updater.cache_partition)), set([
            '5.txt',
            '5.txt.asc',
            '6.txt',
            '6.txt.asc',
            '7.txt',
            '7.txt.asc',
            'device-signing.tar.xz',
            'device-signing.tar.xz.asc',
            'image-master.tar.xz',
            'image-master.tar.xz.asc',
            'image-signing.tar.xz',
            'image-signing.tar.xz.asc',
            'ubuntu_command',
            ]))
        # And the blacklist keyring is available too.
        self.assertEqual(set(os.listdir(config.updater.data_partition)), set([
            'blacklist.tar.xz',
            'blacklist.tar.xz.asc',
            ]))

    @configuration
    def test_reboot(self):
        # Run the intermediate steps, and finish with a reboot.
        self._setup_server_keyrings()
        mediator = Mediator()
        # Mock to check the state of reboot.
        got_reboot = False
        def reboot_mock(self):
            nonlocal got_reboot
            got_reboot = True
        with unittest.mock.patch(
                'systemimage.reboot.Reboot.reboot', reboot_mock):
            mediator.check_for_update()
            mediator.download()
            self.assertFalse(got_reboot)
            mediator.reboot()
            self.assertTrue(got_reboot)

    @configuration
    def test_cancel(self):
        # When we get to the step of downloading the files, cancel it.
        self._setup_server_keyrings()
        mediator = Mediator()
        mediator.check_for_update()
        mediator.cancel()
        self.assertRaises(Canceled, mediator.download)

    @configuration
    def test_callback(self):
        # When downloading, we get callbacks.
        self._setup_server_keyrings()
        received_bytes = 0
        total_bytes = 0
        def callback(received, total):
            nonlocal received_bytes, total_bytes
            received_bytes = received
            total_bytes = total
        mediator = Mediator(callback)
        mediator.check_for_update()
        # Checking for updates does not trigger the callback.
        self.assertEqual(received_bytes, 0)
        self.assertEqual(total_bytes, 0)
        mediator.download()
        # We don't know exactly how many bytes got downloaded, but we know
        # some did.
        self.assertNotEqual(received_bytes, 0)
        self.assertNotEqual(total_bytes, 0)

    @configuration
    def test_pause_resume(self):
        # Pause and resume the download.
        self._setup_server_keyrings()
        for path in ('3/4/5.txt', '4/5/6.txt', '5/6/7.txt'):
            full_path = os.path.join(self._serverdir, path)
            with open(full_path, 'wb') as fp:
                fp.write(b'x' * 100 * MiB)
        mediator = Mediator()
        pauses = []
        def do_paused(self, signal, path, paused):
            if paused:
                pauses.append(datetime.now())
        resumes = []
        def do_resumed(self, signal, path, resumed):
            if resumed:
                resumes.append(datetime.now())
        def pause_on_start(self, signal, path, started):
            if started and self._pausable:
                mediator.pause()
                GLib.timeout_add_seconds(3, mediator.resume)
        with ExitStack() as resources:
            resources.enter_context(
                patch('systemimage.download.DownloadReactor._do_paused',
                      do_paused))
            resources.enter_context(
                patch('systemimage.download.DownloadReactor._do_resumed',
                      do_resumed))
            resources.enter_context(
                patch('systemimage.download.DownloadReactor._do_started',
                      pause_on_start))
            mediator.check_for_update()
            # We'll get a signature error because we messed with the file
            # contents.  Since this check happens after all files are
            # downloaded, this exception is inconsequential to the thing we're
            # testing.
            try:
                mediator.download()
            except SignatureError:
                pass
        # There should be at one pause and one resume event, separated by
        # about 3 seconds.
        self.assertEqual(len(pauses), 1)
        self.assertEqual(len(resumes), 1)
        self.assertGreaterEqual(resumes[0] - pauses[0], timedelta(seconds=2.5))

    @configuration
    def test_state_machine_exceptions(self, ini_file):
        # An exception in the state machine captures the exception and returns
        # an error string in the Update instance.
        self._setup_server_keyrings()
        config = Configuration()
        config.load(ini_file)
        with chmod(config.updater.cache_partition, 0):
            update = Mediator().check_for_update()
        # There's no winning path, but there is an error.
        self.assertFalse(update.is_available)
        self.assertIn('Permission denied', update.error)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Persistent settings - used by the DBus API."""

__all__ = [
    'Settings',
    ]


import sqlite3

from contextlib import contextmanager
from systemimage.config import config

SCHEMA_VERSION = '1'
AUTO_DOWNLOAD_DEFAULT = '1'


class Settings:
    def __init__(self, use_config=None):
        self._use_config = use_config
        # If the database file does not yet exist, create it.
        with self._cursor() as c:
            c.execute('select tbl_name from sqlite_master')
            if len(c.fetchall()) == 0:
                # The database file has no tables.
                c.execute('create table settings (key, value)''')
            # Hopefully we won't ever need to migrate this schema, but just in
            # case we do, set a version value.
            c.execute('insert into settings values ("__version__", ?)',
                      (SCHEMA_VERSION,))

    @contextmanager
    def _cursor(self):
        dbpath = (config.system.settings_db
                  if self._use_config is None
                  else self._use_config.system.settings_db)
        with sqlite3.connect(dbpath) as conn:
            yield conn.cursor()

    def set(self, key, value):
        with self._cursor() as c:
            c.execute('select value from settings where key = ?', (key,))
            row = c.fetchone()
            if row is None:
                c.execute('insert into settings values (?, ?)',
                          (key, value))
            else:
                c.execute('update settings set value = ? where key = ?',
                          (value, key))

    def get(self, key):
        with self._cursor() as c:
            c.execute('select value from settings where key = ?', (key,))
            row = c.fetchone()
            if row is None:
                if key == 'auto_download':
                    return AUTO_DOWNLOAD_DEFAULT
                return ''
            return row[0]



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Various and sundry helpers."""

__all__ = [
    'DEFAULT_DIRMODE',
    'ExtendedEncoder',
    'as_loglevel',
    'as_object',
    'as_timedelta',
    'atomic',
    'last_update_date',
    'makedirs',
    'phased_percentage',
    'safe_remove',
    'temporary_directory',
    'version_detail',
    ]


import os
import re
import json
import time
import random
import shutil
import logging
import tempfile

from contextlib import ExitStack, contextmanager
from datetime import datetime, timedelta
from importlib import import_module
from systemimage.bag import Bag


LAST_UPDATE_FILE = '/userdata/.last_update'
UNIQUE_MACHINE_ID_FILE = '/var/lib/dbus/machine-id'
DEFAULT_DIRMODE = 0o02700


def safe_remove(path):
    """Like os.remove() but don't complain if the file doesn't exist."""
    try:
        os.remove(path)
    except (FileNotFoundError, IsADirectoryError):
        pass


@contextmanager
def atomic(dst, encoding='utf-8'):
    """Open a temporary file for writing using the given encoding.

    The context manager returns an open file object, into which you can write
    text or bytes depending on the encoding it was opened with.  Upon exit,
    the temporary file is moved atomically to the destination.  If an
    exception occurs, the temporary file is removed.

    :param dst: The path name of the target file.
    :param encoding: The encoding to use for the open file.  If None, then
        file is opened in binary mode.
    """
    directory = os.path.dirname(dst)
    fd, temp = tempfile.mkstemp(dir=directory)
    with ExitStack() as stack:
        stack.callback(safe_remove, temp)
        os.close(fd)
        mode = 'wb' if encoding is None else 'wt'
        with open(temp, mode, encoding=encoding) as fp:
            yield fp
        os.rename(temp, dst)


# This is stolen directly out of lazr.config.  We can do that since we own
# both code bases. :)
def _sortkey(item):
    """Return a value that sorted(..., key=_sortkey) can use."""
    order = dict(
        w=0,    # weeks
        d=1,    # days
        h=2,    # hours
        m=3,    # minutes
        s=4,    # seconds
        )
    return order.get(item[-1])


class _Called:
    # Defer importing named object until it's actually called.  This should
    # reduce the instances of circular imports.
    def __init__(self, path):
        self._path, dot, self._name = path.rpartition('.')
        if dot != '.':
            raise ValueError

    def _dig(self):
        module = import_module(self._path)
        return getattr(module, self._name)

    def __call__(self, *args, **kws):
        return self._dig()(*args, **kws)

    def __eq__(self, other):
        # Let class equality (and in-equality) work.
        myself = self._dig()
        return myself == other

    def __ne__(self, other):
        return not self.__eq__(other)


def as_object(value):
    """Convert a Python dotted-path specification to an object.

    :param value: A dotted-path specification,
        e.g. the string `systemimage.scores.WeightedScorer`
    :return: A proxy object that when called, performs the import and calls
        the underyling object.
    :raises ValueError: when `value` is not dotted.
    """
    return _Called(value)


def as_timedelta(value):
    """Convert a value string to the equivalent timedelta."""
    # Technically, the regex will match multiple decimal points in the
    # left-hand side, but that's okay because the float/int conversion below
    # will properly complain if there's more than one dot.
    components = sorted(re.findall(r'([\d.]+[smhdw])', value), key=_sortkey)
    # Complain if the components are out of order.
    if ''.join(components) != value:
        raise ValueError
    keywords = dict((interval[0].lower(), interval)
                    for interval in ('weeks', 'days', 'hours',
                                     'minutes', 'seconds'))
    keyword_arguments = {}
    for interval in components:
        if len(interval) == 0:
            raise ValueError
        keyword = keywords.get(interval[-1].lower())
        if keyword is None:
            raise ValueError
        if keyword in keyword_arguments:
            raise ValueError
        if '.' in interval[:-1]:
            converted = float(interval[:-1])
        else:
            converted = int(interval[:-1])
        keyword_arguments[keyword] = converted
    if len(keyword_arguments) == 0:
        raise ValueError
    return timedelta(**keyword_arguments)


def as_loglevel(value):
    level = getattr(logging, value.upper(), None)
    if level is None or not isinstance(level, int):
        raise ValueError
    return level


class ExtendedEncoder(json.JSONEncoder):
    """An extended JSON encoder which knows about other data types."""

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            # as_timedelta() does not recognize microseconds, so convert these
            # to floating seconds, but only if there are any seconds.
            if obj.seconds > 0 or obj.microseconds > 0:
                seconds = obj.seconds + obj.microseconds / 1000000.0
                return '{0}d{1}s'.format(obj.days, seconds)
            return '{0}d'.format(obj.days)
        elif isinstance(obj, Bag):
            return obj.original
        return json.JSONEncoder.default(self, obj)


@contextmanager
def temporary_directory(*args, **kws):
    """A context manager that creates a temporary directory.

    The directory and all its contents are deleted when the context manager
    exits.  All positional and keyword arguments are passed to mkdtemp().
    """
    tempdir = tempfile.mkdtemp(*args, **kws)
    os.chmod(tempdir, kws.get('mode', DEFAULT_DIRMODE))
    try:
        yield tempdir
    finally:
        try:
            shutil.rmtree(tempdir)
        except FileNotFoundError:
            pass


def makedirs(dir, mode=DEFAULT_DIRMODE):
    try:
        os.makedirs(dir, mode=mode, exist_ok=True)
    except (PermissionError, FileExistsError):
        pass


def last_update_date():
    """Return the last update date.

    Taken from the mtime of the following files, in order:

    - /userdata/.last_update
    - /etc/system-image/channel.ini
    - /etc/ubuntu-build

    First existing path wins.
    """
    # Avoid circular imports.
    from systemimage.config import config
    channel_ini = os.path.join(
        os.path.dirname(config.config_file), 'channel.ini')
    ubuntu_build = config.system.build_file
    for path in (LAST_UPDATE_FILE, channel_ini, ubuntu_build):
        try:
            # Local time, since we can't know the timezone.
            timestamp = datetime.fromtimestamp(os.stat(path).st_mtime)
            # Seconds resolution.
            timestamp = timestamp.replace(microsecond=0)
            return str(timestamp)
        except FileNotFoundError:
            pass
    else:
        return 'Unknown'


def version_detail():
    """Return a dictionary of the version details."""
    # Avoid circular imports.
    from systemimage.config import config
    version_details = getattr(config.service, 'version_detail', None)
    if version_details is None:
        return {}
    details = {}
    if version_details is not None:
        for item in version_details.strip().split(','):
            name, equals, version = item.partition('=')
            if equals != '=':
                continue
            details[name] = version
    return details


_pp_cache = None

def phased_percentage(*, reset=False):
    global _pp_cache
    if _pp_cache is None:
        with open(UNIQUE_MACHINE_ID_FILE, 'rb') as fp:
            data = fp.read()
        now = str(time.time()).encode('us-ascii')
        r = random.Random()
        r.seed(data + now)
        _pp_cache = r.randint(0, 100)
    try:
        return _pp_cache
    finally:
        if reset:
            _pp_cache = None



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the SystemImage dbus service."""

__all__ = [
    'TestDBusApply',
    'TestDBusCheckForUpdate',
    'TestDBusCheckForUpdateException',
    'TestDBusClient',
    'TestDBusDownload',
    'TestDBusGetSet',
    'TestDBusInfo',
    'TestDBusInfoNoDetails',
    'TestDBusMockFailApply',
    'TestDBusMockFailPause',
    'TestDBusMockFailResume',
    'TestDBusMockNoUpdate',
    'TestDBusMockUpdateAutoSuccess',
    'TestDBusMockUpdateManualSuccess',
    'TestDBusMultipleChecksInFlight',
    'TestDBusPauseResume',
    'TestDBusProgress',
    'TestDBusRegressions',
    'TestDBusUseCache',
    ]


import os
import dbus
import time
import shutil
import unittest

from contextlib import ExitStack
from datetime import datetime
from dbus.exceptions import DBusException
from functools import partial
from systemimage.bindings import DBusClient
from systemimage.config import Configuration
from systemimage.helpers import atomic, safe_remove
from systemimage.reactor import Reactor
from systemimage.testing.helpers import (
    copy, find_dbus_process, make_http_server, setup_index, setup_keyring_txz,
    setup_keyrings, sign)
from systemimage.testing.nose import SystemImagePlugin


class SignalCapturingReactor(Reactor):
    def __init__(self, *signals):
        super().__init__(dbus.SystemBus())
        for signal in signals:
            self.react_to(signal)
        self.signals = []

    def _default(self, signal, path, *args, **kws):
        self.signals.append(args)
        self.quit()

    def run(self, method=None, timeout=None):
        if method is not None:
            self.schedule(method)
        super().run(timeout=timeout)


class AutoDownloadCancelingReactor(Reactor):
    def __init__(self, iface):
        super().__init__(dbus.SystemBus())
        self._iface = iface
        self.got_update_available_status = False
        self.got_update_failed = False
        self.react_to('UpdateAvailableStatus')
        self.react_to('UpdateFailed')

    def _do_UpdateAvailableStatus(self, signal, path, *args, **kws):
        self.got_update_available_status = True
        self._iface.CancelUpdate()

    def _do_UpdateFailed(self, signal, path, *args, **kws):
        self.got_update_failed = True
        self.quit()


class ProgressRecordingReactor(Reactor):
    def __init__(self):
        super().__init__(dbus.SystemBus())
        self.react_to('UpdateDownloaded')
        self.react_to('UpdateProgress')
        self.progress = []

    def _do_UpdateDownloaded(self, signal, path, *args, **kws):
        self.quit()

    def _do_UpdateProgress(self, signal, path, *args, **kws):
        self.progress.append(args)


class PausingReactor(Reactor):
    def __init__(self, iface):
        super().__init__(dbus.SystemBus())
        self._iface = iface
        self.paused = False
        self.react_to('UpdateProgress')
        self.react_to('UpdatePaused')

    def _do_UpdateProgress(self, signal, path, percentage, eta):
        if percentage == 0:
            self._iface.PauseDownload()

    def _do_UpdatePaused(self, signal, path, percentage):
        self.paused = True
        self.quit()


class DoubleCheckingReactor(Reactor):
    def __init__(self, iface):
        super().__init__(dbus.SystemBus())
        self.iface = iface
        self.uas_signals = []
        self.react_to('UpdateAvailableStatus')
        self.react_to('UpdateDownloaded')
        self.schedule(self.iface.CheckForUpdate)

    def _do_UpdateAvailableStatus(self, signal, path, *args, **kws):
        # We'll keep doing this until we get the UpdateDownloaded signal.
        self.uas_signals.append(args)
        self.schedule(self.iface.CheckForUpdate)

    def _do_UpdateDownloaded(self, *args, **kws):
        self.quit()


class ManualUpdateReactor(Reactor):
    def __init__(self, iface):
        super().__init__(dbus.SystemBus())
        self.iface = iface
        self.rebooting = False
        self.react_to('UpdateAvailableStatus')
        self.react_to('UpdateProgress')
        self.react_to('UpdateDownloaded')
        self.react_to('Rebooting')
        self.react_to('UpdateFailed')
        self.iface.CheckForUpdate()

    def _do_UpdateAvailableStatus(self, signal, path, *args, **kws):
        # When the update is available, start the download.
        self.iface.DownloadUpdate()

    def _do_UpdateProgress(self, signal, path, *args, **kws):
        # Once the download is in progress, initiate another check.  Only do
        # this on the first progress signal.
        if args == (0, 0):
            self.iface.CheckForUpdate()

    def _do_UpdateDownloaded(self, signal, path, *args, **kws):
        # The update successfully downloaded, so apply the update now.
        self.iface.ApplyUpdate()

    def _do_UpdateFailed(self, signal, path, *args, **kws):
        # Before LP: #1287919 was fixed, this signal would have been sent.
        self.rebooting = False
        self.quit()

    def _do_Rebooting(self, signal, path, *args, **kws):
        # The system is now rebooting <wink>.
        self.rebooting = True
        self.quit()


class _TestBase(unittest.TestCase):
    """Base class for all DBus testing."""

    # For unittest's assertMultiLineEqual().
    maxDiff = None

    # Override this to start the DBus server in a different testing mode.
    mode = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        SystemImagePlugin.controller.set_mode(
            cert_pem='cert.pem',
            service_mode=cls.mode)

    @classmethod
    def tearDownClass(cls):
        # Try to find the system-image-dbus process matching the current ini
        # file used by the controller.  We'll use this to ensure that the
        # process has exited.
        process = find_dbus_process(SystemImagePlugin.controller.ini_path)
        if process is not None:
            bus = dbus.SystemBus()
            service = bus.get_object('com.canonical.SystemImage', '/Service')
            iface = dbus.Interface(service, 'com.canonical.SystemImage')
            iface.Exit()
            # Wait for the process to exit, but only for a little while.
            process.wait(60)
        # Clear out the temporary directory.
        config = Configuration()
        config.load(SystemImagePlugin.controller.ini_path)
        try:
            shutil.rmtree(config.system.tempdir)
        except FileNotFoundError:
            pass
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.system_bus = dbus.SystemBus()
        service = self.system_bus.get_object(
            'com.canonical.SystemImage', '/Service')
        self.iface = dbus.Interface(service, 'com.canonical.SystemImage')

    def tearDown(self):
        self.iface.Reset()
        super().tearDown()

    def download_manually(self):
        self.iface.SetSetting('auto_download', '0')

    def download_on_wifi(self):
        self.iface.SetSetting('auto_download', '1')

    def download_always(self):
        self.iface.SetSetting('auto_download', '2')


class _LiveTesting(_TestBase):
    mode = 'live'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._resources = ExitStack()
        # Set up the http/https servers that the dbus client will talk to.
        # Start up both an HTTPS and HTTP server.  The data files are vended
        # over the latter, everything else, over the former.
        serverdir = SystemImagePlugin.controller.serverdir
        try:
            cls._resources.push(
                make_http_server(serverdir, 8943, 'cert.pem', 'key.pem'))
            cls._resources.push(make_http_server(serverdir, 8980))
            # Set up the server files.
            copy('channels_06.json', serverdir, 'channels.json')
            sign(os.path.join(serverdir, 'channels.json'), 'image-signing.gpg')
            # Only the archive-master key is pre-loaded.  All the other keys
            # are downloaded and there will be both a blacklist and device
            # keyring.  The four signed keyring tar.xz files and their
            # signatures end up in the proper location after the state machine
            # runs to completion.
            config = Configuration()
            config.load(SystemImagePlugin.controller.ini_path)
            setup_keyrings('archive-master', use_config=config)
            setup_keyring_txz(
                'spare.gpg', 'image-master.gpg', dict(type='blacklist'),
                os.path.join(serverdir, 'gpg', 'blacklist.tar.xz'))
            setup_keyring_txz(
                'image-master.gpg', 'archive-master.gpg',
                dict(type='image-master'),
                os.path.join(serverdir, 'gpg', 'image-master.tar.xz'))
            setup_keyring_txz(
                'image-signing.gpg', 'image-master.gpg',
                dict(type='image-signing'),
                os.path.join(serverdir, 'gpg', 'image-signing.tar.xz'))
            setup_keyring_txz(
                'device-signing.gpg', 'image-signing.gpg',
                dict(type='device-signing'),
                os.path.join(serverdir, 'stable', 'nexus7',
                             'device-signing.tar.xz'))
        except:
            cls._resources.close()
            raise

    @classmethod
    def tearDownClass(cls):
        cls._resources.close()
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self._prepare_index('index_13.json')
        # We need a configuration file that agrees with the dbus client.
        self.config = Configuration()
        self.config.load(SystemImagePlugin.controller.ini_path)
        # For testing reboot preparation.
        self.command_file = os.path.join(
            self.config.updater.cache_partition, 'ubuntu_command')
        # For testing the reboot command without actually rebooting.
        self.reboot_log = os.path.join(
            self.config.updater.cache_partition, 'reboot.log')

    def tearDown(self):
        # Consume the UpdateFailed that results from the cancellation.
        reactor = SignalCapturingReactor('TornDown')
        reactor.run(self.iface.TearDown, timeout=15)
        safe_remove(self.config.system.build_file)
        for updater_dir in (self.config.updater.cache_partition,
                            self.config.updater.data_partition):
            try:
                all_files = os.listdir(updater_dir)
            except FileNotFoundError:
                # The directory itself may not exist.
                pass
            for filename in all_files:
                safe_remove(os.path.join(updater_dir, filename))
        safe_remove(self.reboot_log)
        super().tearDown()

    def _prepare_index(self, index_file, write_callback=None):
        serverdir = SystemImagePlugin.controller.serverdir
        index_path = os.path.join(serverdir, 'stable', 'nexus7', 'index.json')
        head, tail = os.path.split(index_path)
        copy(index_file, head, tail)
        sign(index_path, 'device-signing.gpg')
        setup_index(index_file, serverdir, 'device-signing.gpg',
                    write_callback)

    def _touch_build(self, version):
        # Unlike the touch_build() helper, this one uses our own config object
        # rather than the global one.   It's not worth messing with
        # touch_build() to generalize it.
        assert 0 <= version < (1 << 16), (
            'old style version number: {}'.format(version))
        with open(self.config.system.build_file, 'w', encoding='utf-8') as fp:
            print(version, file=fp)


class TestDBusCheckForUpdate(_LiveTesting):
    """Test the SystemImage dbus service."""

    def setUp(self):
        super().setUp()
        self._more_resources = ExitStack()

    def tearDown(self):
        self._more_resources.close()
        super().tearDown()

    def test_update_available(self):
        # There is an update available.
        self.download_manually()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertEqual(available_version, '1600')
        self.assertEqual(update_size, 314572800)
        # This is the first update applied.
        self.assertEqual(last_update_date, 'Unknown')
        ## self.assertEqual(descriptions, [{'description': 'Full'}])
        self.assertEqual(error_reason, '')

    def test_update_available_auto_download(self):
        # Automatically download the available update.
        self.download_always()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         # descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertTrue(downloading)
        self.assertEqual(available_version, '1600')
        self.assertEqual(update_size, 314572800)
        # This is the first update applied.
        self.assertEqual(last_update_date, 'Unknown')
        ## self.assertEqual(descriptions, [{'description': 'Full'}])
        self.assertEqual(error_reason, '')

    def test_no_update_available(self):
        # Our device is newer than the version that's available.
        self._touch_build(1701)
        # Give /etc/ubuntu-build a predictable mtime.
        timestamp = int(datetime(2013, 8, 1, 10, 11, 12).timestamp())
        os.utime(self.config.system.build_file, (timestamp, timestamp))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertFalse(is_available)
        # No update has been previously applied.
        self.assertEqual(last_update_date, '2013-08-01 10:11:12')
        # All other values are undefined.

    def test_last_update_date(self):
        # Pretend the device got a previous update.  Now, there's no update
        # available, but the date of the last update is provided in the signal.
        self._touch_build(1701)
        # Fake that there was a previous update.
        timestamp = int(datetime(2013, 1, 20, 12, 1, 45).timestamp())
        channel_ini = os.path.join(
            os.path.dirname(SystemImagePlugin.controller.ini_path),
            'channel.ini')
        self._more_resources.callback(safe_remove, channel_ini)
        with open(channel_ini, 'w', encoding='utf-8'):
            pass
        os.utime(channel_ini, (timestamp, timestamp))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertFalse(is_available)
        # No update has been previously applied.
        self.assertEqual(last_update_date, '2013-01-20 12:01:45')
        # All other values are undefined.

    def test_check_for_update_twice(self):
        # Issue two CheckForUpdate calls immediate after each other.
        self.download_always()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        def two_calls():
            self.iface.CheckForUpdate()
            self.iface.CheckForUpdate()
        reactor.run(two_calls)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         # descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertTrue(downloading)

    @unittest.skip('LP: #1215586')
    def test_get_multilingual_descriptions(self):
        # The descriptions are multilingual.
        self._prepare_index('index_14.json')
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date, descriptions, error_reason) = reactor.signals[0]
        self.assertEqual(descriptions, [
            {'description': 'Full B',
             'description-en': 'The full B',
            },
            {'description': 'Delta B.1',
             'description-en_US': 'This is the delta B.1',
             'description-xx': 'XX This is the delta B.1',
             'description-yy': 'YY This is the delta B.1',
             'description-yy_ZZ': 'YY-ZZ This is the delta B.1',
            },
            {'description': 'Delta B.2',
             'description-xx': 'Oh delta, my delta',
             'description-xx_CC': 'This hyar is the delta B.2',
            }])


class TestDBusDownload(_LiveTesting):
    def test_auto_download(self):
        # When always auto-downloading, and there is an update available, the
        # update gets automatically downloaded.  First, we'll get an
        # UpdateAvailableStatus signal, followed by a bunch of UpdateProgress
        # signals (which for this test, we'll ignore), and finally an
        # UpdateDownloaded signal.
        self.download_always()
        self.assertFalse(os.path.exists(self.command_file))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertTrue(downloading)
        # Now, wait for the UpdateDownloaded signal.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run()
        self.assertEqual(len(reactor.signals), 1)
        with open(self.command_file, 'r', encoding='utf-8') as fp:
            command = fp.read()
        self.assertMultiLineEqual(command, """\
load_keyring image-master.tar.xz image-master.tar.xz.asc
load_keyring image-signing.tar.xz image-signing.tar.xz.asc
load_keyring device-signing.tar.xz device-signing.tar.xz.asc
format system
mount system
update 6.txt 6.txt.asc
update 7.txt 7.txt.asc
update 5.txt 5.txt.asc
unmount system
""")

    def test_nothing_to_auto_download(self):
        # We're auto-downloading, but there's no update available.
        self.download_always()
        self._touch_build(1701)
        self.assertFalse(os.path.exists(self.command_file))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertFalse(is_available)
        self.assertFalse(downloading)
        # Now, wait for the UpdateDownloaded signal, which isn't coming.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        self.assertFalse(os.path.exists(self.command_file))

    def test_manual_download(self):
        # When manually downloading, and there is an update available, the
        # update does not get downloaded until we explicitly ask it to be.
        self.download_manually()
        self.assertFalse(os.path.exists(self.command_file))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        # This is false because we're in manual download mode.
        self.assertFalse(downloading)
        self.assertFalse(os.path.exists(self.command_file))
        # No UpdateDownloaded signal is coming.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        self.assertFalse(os.path.exists(self.command_file))
        # Now we download manually and wait again for the signal.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(self.iface.DownloadUpdate)
        with open(self.command_file, 'r', encoding='utf-8') as fp:
            command = fp.read()
        self.assertMultiLineEqual(command, """\
load_keyring image-master.tar.xz image-master.tar.xz.asc
load_keyring image-signing.tar.xz image-signing.tar.xz.asc
load_keyring device-signing.tar.xz device-signing.tar.xz.asc
format system
mount system
update 6.txt 6.txt.asc
update 7.txt 7.txt.asc
update 5.txt 5.txt.asc
unmount system
""")

    def test_nothing_to_manually_download(self):
        # We're manually downloading, but there's no update available.
        self.download_manually()
        self._touch_build(1701)
        self.assertFalse(os.path.exists(self.command_file))
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertFalse(is_available)
        # This is false because we're in manual download mode.
        self.assertFalse(downloading)
        # No UpdateDownloaded signal is coming
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        self.assertFalse(os.path.exists(self.command_file))
        # Now we download manually, but no signal is coming.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(self.iface.DownloadUpdate, timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        self.assertFalse(os.path.exists(self.command_file))

    def test_update_failed_signal(self):
        # A signal is issued when the update failed.
        self.download_manually()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        # Cause the update to fail by deleting a file from the server.
        os.remove(os.path.join(SystemImagePlugin.controller.serverdir,
                               '4/5/6.txt.asc'))
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        failure_count, last_reason = reactor.signals[0]
        self.assertEqual(failure_count, 1)
        self.assertEqual(last_reason[:17], 'FileNotFoundError')

    def test_duplicate_destinations(self):
        # A faulty index.json might specify that two different urls yield the
        # same local destination file.  This is a bug on the server and the
        # client cannot perform an update.
        self.download_manually()
        self._prepare_index('index_23.json')
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        failure_count, last_reason = reactor.signals[0]
        self.assertEqual(failure_count, 1)
        # Don't count on a specific error message.
        self.assertEqual(
            last_reason[:46], 'systemimage.download.DuplicateDestinationError')


class TestDBusApply(_LiveTesting):
    def setUp(self):
        super().setUp()
        self.download_always()

    def test_reboot(self):
        # Apply the update, which reboots the device.
        self.assertFalse(os.path.exists(self.reboot_log))
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(self.iface.CheckForUpdate)
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertTrue(reactor.signals[0])
        with open(self.reboot_log, encoding='utf-8') as fp:
            reboot = fp.read()
        self.assertEqual(reboot, '/sbin/reboot -f recovery')

    def test_reboot_no_update(self):
        # There's no update to reboot to.
        self.assertFalse(os.path.exists(self.reboot_log))
        self._touch_build(1701)
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertFalse(reactor.signals[0][0])
        self.assertFalse(os.path.exists(self.reboot_log))

    def test_reboot_after_update_failed(self):
        # Cause the update to fail by deleting a file from the server.
        self.download_manually()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        os.remove(os.path.join(SystemImagePlugin.controller.serverdir,
                               '4/5/6.txt.asc'))
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        failure_count, reason = reactor.signals[0]
        self.assertEqual(failure_count, 1)
        self.assertNotEqual(reason, '')
        # The reboot fails, so we get an error message.
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertFalse(reactor.signals[0][0])

    def test_auto_download_cancel(self):
        # While automatically downloading, cancel the update.
        self.download_always()
        reactor = AutoDownloadCancelingReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        self.assertTrue(reactor.got_update_available_status)
        self.assertTrue(reactor.got_update_failed)

    def test_exit(self):
        # There is a D-Bus method to exit the server immediately.
        self.iface.Exit()
        self.assertRaises(DBusException, self.iface.Info)
        # Re-establish a new connection.
        bus = dbus.SystemBus()
        service = bus.get_object('com.canonical.SystemImage', '/Service')
        self.iface = dbus.Interface(service, 'com.canonical.SystemImage')
        # There's no update to apply, so we'll get an error string instead of
        # the empty string for this call.  But it will restart the server.
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertFalse(reactor.signals[0][0])

    def test_cancel_while_not_downloading(self):
        # If we call CancelUpdate() when we're not downloading anything, no
        # UpdateFailed signal is sent.
        self.download_manually()
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # Since we're downloading manually, no signal will be sent.
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.CancelUpdate, timeout=15)
        self.assertEqual(len(reactor.signals), 0)

    def test_cancel_manual(self):
        # While manually downloading, cancel the update.
        self.download_manually()
        # The downloads can be canceled when there is an update available.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        # Cancel future operations.  However, since no download is in
        # progress, we will not get a signal.
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.CancelUpdate, timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        self.assertFalse(os.path.exists(self.command_file))
        # Try to download the update again, though this will fail again.
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        failure_count, reason = reactor.signals[0]
        self.assertEqual(failure_count, 2)
        self.assertNotEqual(reason, '')
        self.assertFalse(os.path.exists(self.command_file))
        # The next check resets the failure count and succeeds.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        # And now we can successfully download the update.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)


class MockReactor(Reactor):
    def __init__(self, iface):
        super().__init__(dbus.SystemBus())
        self._iface = iface
        self.timeout = 120
        self.react_to('UpdateProgress')
        self.pause_at_percentage = None
        self.cancel_at_percentage = None
        self.pause_start = None
        self.pause_end = None
        self.progress = []
        self.react_to('UpdateDownloaded')
        self.downloaded = False
        self.react_to('UpdateAvailableStatus')
        self.status = None
        self.auto_download = True
        self.react_to('UpdateFailed')
        self.failed = []
        self.react_to('UpdatePaused')
        self.pauses = []
        self.pauses_should_quit = True

    def _resume(self):
        self.pause_end = time.time()
        self._iface.DownloadUpdate()
        return False

    def _do_UpdateProgress(self, signal, path, percentage, eta):
        self.progress.append((percentage, eta))
        if percentage == self.pause_at_percentage:
            self.pause_start = time.time()
            self._iface.PauseDownload()
            # Wait 5 seconds, then resume the download.
            self.schedule(self._resume, 5000)
        elif percentage == self.cancel_at_percentage:
            self._iface.CancelUpdate()

    def _do_UpdateDownloaded(self, *args, **kws):
        self.downloaded = True
        self.quit()

    def _do_UpdateAvailableStatus(self, signal, path, *args, **kws):
        self.status = args
        if not self.auto_download:
            # The download must be started manually.
            self.quit()

    def _do_UpdateFailed(self, signal, path, *args, **kws):
        self.failed.append(args)
        self.quit()

    def _do_UpdatePaused(self, signal, path, percentage):
        self.pauses.append(percentage)
        if self.pauses_should_quit:
            self.quit()


class TestDBusMockUpdateAutoSuccess(_TestBase):
    mode = 'update-auto-success'

    def test_scenario_1(self):
        # Start the ball rolling.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertTrue(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, '')
        # We should have gotten 100 UpdateProgress signals, where each
        # increments the percentage by 1 and decrements the eta by 0.5.
        self.assertEqual(len(reactor.progress), 100)
        for i in range(100):
            percentage, eta = reactor.progress[i]
            self.assertEqual(percentage, i)
            self.assertEqual(eta, 50 - (i * 0.5))
        self.assertTrue(reactor.downloaded)
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertTrue(reactor.signals[0])

    def test_scenario_2(self):
        # Like scenario 1, but with PauseDownload called during the downloads.
        reactor = MockReactor(self.iface)
        reactor.pauses_should_quit = False
        reactor.pause_at_percentage = 35
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        # We got a pause signal.
        self.assertEqual(len(reactor.pauses), 1)
        self.assertEqual(reactor.pauses[0], 35)
        # Make sure that we still got 100 progress reports.
        self.assertEqual(len(reactor.progress), 100)
        # And we still completed successfully.
        self.assertTrue(reactor.downloaded)
        # And that we paused successfully.  We can't be exact about the amount
        # of time we paused, but it should always be at least 4 seconds.
        self.assertGreater(reactor.pause_end - reactor.pause_start, 4)

    def test_scenario_3(self):
        # Like scenario 2, but PauseDownload is called when not downloading,
        # so it is a no-op.  The test service waits 3 seconds after a
        # CheckForUpdate before it begins downloading, so let's issue a
        # no-op PauseDownload after 1 second.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.PauseDownload, 1000)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        self.assertEqual(len(reactor.pauses), 0)
        self.assertIsNone(reactor.pause_start)
        self.assertIsNone(reactor.pause_end)

    def test_scenario_4(self):
        # If DownloadUpdate is called when not paused, downloading, or
        # update-checked, it is a no-op.
        self.iface.DownloadUpdate()
        # Only run for 15 seconds, but still, we'll never see an
        # UpdateAvailableStatus or UpdateDownloaded.
        reactor = MockReactor(self.iface)
        reactor.timeout = 15
        reactor.run()
        self.assertIsNone(reactor.status)
        self.assertFalse(reactor.downloaded)

    def test_scenario_5(self):
        # In this scenario, we cancel the download midway through.  This will
        # result in an UpdateFailed signal.
        reactor = MockReactor(self.iface)
        reactor.cancel_at_percentage = 27
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        # Our failed signal will tell us we got one consecutive failure and
        # the reason is that we canceled (but don't depend on the exact
        # content of the last_reason, just that it's not the empty string).
        self.assertEqual(len(reactor.failed), 1)
        failure_count, reason = reactor.failed[0]
        self.assertEqual(failure_count, 1)
        self.assertNotEqual(reason, '')
        # We also didn't download anything.
        self.assertFalse(reactor.downloaded)

    def test_scenario_6(self):
        # Like secenario 5, but after a cancel, CheckForUpdate will restart
        # things again.
        reactor = MockReactor(self.iface)
        reactor.cancel_at_percentage = 13
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        # Our failed signal will tell us we got one consecutive failure and
        # the reason is that we canceled (but don't depend on the exact
        # content of the last_reason, just that it's not the empty string).
        self.assertEqual(len(reactor.failed), 1)
        failure_count, reason = reactor.failed[0]
        self.assertEqual(failure_count, 1)
        self.assertNotEqual(reason, '')
        # We also didn't download anything.
        self.assertFalse(reactor.downloaded)
        # Now, restart the download.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        # This time, we've downloaded everything
        self.assertTrue(reactor.downloaded)
        self.assertEqual(len(reactor.failed), 0)


class TestDBusMockUpdateManualSuccess(_TestBase):
    mode = 'update-manual-success'

    def test_scenario_1(self):
        # Like scenario 1 for auto-downloading except that the download must
        # be started explicitly.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.auto_download = False
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, '')
        # There should be no progress yet.
        self.assertEqual(len(reactor.progress), 0)
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.DownloadUpdate)
        reactor.auto_download = False
        reactor.run()
        # We should have gotten 100 UpdateProgress signals, where each
        # increments the percentage by 1 and decrements the eta by 0.5.
        self.assertEqual(len(reactor.progress), 100)
        for i in range(100):
            percentage, eta = reactor.progress[i]
            self.assertEqual(percentage, i)
            self.assertEqual(eta, 50 - (i * 0.5))
        self.assertTrue(reactor.downloaded)
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertTrue(reactor.signals[0])


class TestDBusMockUpdateFailed(_TestBase):
    mode = 'update-failed'

    def test_scenario_1(self):
        # The server is already in falure mode.  A CheckForUpdate() restarts
        # the check, which returns information about the new update.  It
        # auto-starts, but this fails.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, 'You need some network for downloading')
        self.assertEqual(len(reactor.failed), 1)
        failure_count, reason = reactor.failed[0]
        self.assertEqual(failure_count, 2)
        self.assertEqual(reason, 'You need some network for downloading')


class TestDBusMockFailApply(_TestBase):
    mode = 'fail-apply'

    def test_scenario_1(self):
        # The update has been downloaded, client sends CheckForUpdate and
        # receives a response.  The update is downloaded successfully.  An
        # error occurs when we try to apply the update.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, '')
        self.assertTrue(reactor.downloaded)
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertFalse(bool(reactor.signals[0][0]))


class TestDBusMockFailResume(_TestBase):
    mode = 'fail-resume'

    def test_scenario_1(self):
        # The server download is paused at 42%.  A CheckForUpdate is issued
        # and gets a response.  An UpdatePaused signal is sent.  A problem
        # occurs that prevents resuming.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, '')
        # The download is already paused.
        self.assertEqual(len(reactor.pauses), 1)
        self.assertEqual(reactor.pauses[0], 42)
        # We try to resume the download, but that fails.
        self.assertEqual(len(reactor.failed), 0)
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.DownloadUpdate)
        reactor.run()
        self.assertEqual(len(reactor.failed), 1)
        failure_count, reason = reactor.failed[0]
        self.assertEqual(failure_count, 9)
        self.assertEqual(reason, 'You need some network for downloading')


class TestDBusMockFailPause(_TestBase):
    mode = 'fail-pause'

    def test_scenario_1(self):
        # The server is downloading, currently at 10% with no known ETA.  The
        # client tries to pause the download but is unable to do so.
        reactor = MockReactor(self.iface)
        reactor.schedule(self.iface.CheckForUpdate)
        # Only run the loop for a few seconds, since otherwise there's no
        # natural way to pause the download.
        reactor.timeout = 5
        reactor.run()
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.status
        self.assertTrue(is_available)
        self.assertTrue(downloading)
        self.assertEqual(available_version, '42')
        self.assertEqual(update_size, 1337 * 1024 * 1024)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        ## self.assertEqual(descriptions, [
        ##     {'description': 'Ubuntu Edge support',
        ##      'description-en_GB': 'change the background colour',
        ##      'description-fr': "Support d'Ubuntu Edge",
        ##     },
        ##     {'description':
        ##      'Flipped container with 200% boot speed improvement',
        ##     }])
        self.assertEqual(error_reason, '')
        self.assertEqual(len(reactor.progress), 1)
        percentage, eta = reactor.progress[0]
        self.assertEqual(percentage, 10)
        self.assertEqual(eta, 0)
        reason = self.iface.PauseDownload()
        self.assertEqual(reason, 'no no, not now')


class TestDBusMockNoUpdate(_TestBase):
    mode = 'no-update'

    def test_scenario_1(self):
        # No update is available.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertFalse(is_available)
        self.assertFalse(downloading)
        self.assertEqual(last_update_date, '1983-09-13T12:13:14')
        # All the other status variables can be ignored.

    def test_lp_1215946(self):
        reactor = MockReactor(self.iface)
        reactor.auto_download = False
        # no-update mock sends UpdateFailed before UpdateAvailableStatus.
        reactor.schedule(self.iface.CheckForUpdate)
        reactor.run()
        self.assertEqual(len(reactor.failed), 0)
        self.assertIsNotNone(reactor.status)


class TestDBusClient(_LiveTesting):
    """Test the DBus client (used with --dbus)."""

    def setUp(self):
        super().setUp()
        self._client = DBusClient()

    def test_check_for_update(self):
        # There is an update available.
        self._client.check_for_update()
        self.assertTrue(self._client.is_available)
        self.assertTrue(self._client.downloaded)

    def test_check_for_no_update(self):
        # There is no update available.
        self._touch_build(1701)
        self._client.check_for_update()
        self.assertFalse(self._client.is_available)
        self.assertFalse(self._client.downloaded)

    def test_update_failed(self):
        # For some reason <wink>, the update fails.
        #
        # Cause the update to fail by deleting a file from the server.
        os.remove(os.path.join(SystemImagePlugin.controller.serverdir,
                               '4/5/6.txt.asc'))
        self._client.check_for_update()
        self.assertTrue(self._client.is_available)
        self.assertFalse(self._client.downloaded)
        self.assertTrue(self._client.failed)

    def test_reboot(self):
        # After a successful update, we can reboot.
        self._client.check_for_update()
        self.assertTrue(self._client.is_available)
        self.assertTrue(self._client.downloaded)
        self._client.reboot()
        with open(self.reboot_log, encoding='utf-8') as fp:
            reboot = fp.read()
        self.assertEqual(reboot, '/sbin/reboot -f recovery')


class TestDBusRegressions(_LiveTesting):
    """Test that various regressions have been fixed."""

    def test_lp_1205398(self):
        # Reset state after cancel.
        self.download_manually()
        # This test requires that the download take more than 50ms, since
        # that's the quickest we can issue the cancel, so make one of the
        # files huge.
        serverdir = SystemImagePlugin.controller.serverdir
        index_path = os.path.join(serverdir, 'stable', 'nexus7', 'index.json')
        file_path = os.path.join(serverdir, '5', '6', '7.txt')
        # This index file has a 5/6/7/txt checksum equal to the one we're
        # going to create below.
        setup_index('index_18.json', serverdir, 'device-signing.gpg')
        head, tail = os.path.split(index_path)
        copy('index_18.json', head, tail)
        sign(index_path, 'device-signing.gpg')
        with open(file_path, 'wb') as fp:
            # 50MB
            for chunk in range(12800):
                fp.write(b'x' * 4096)
        sign(file_path, 'device-signing.gpg')
        # An update is available.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        self.assertFalse(os.path.exists(self.command_file))
        # Arrange for the download to be canceled after it starts.
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.schedule(self.iface.CancelUpdate)
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        failure_count, reason = reactor.signals[0]
        self.assertNotEqual(reason, '')
        self.assertFalse(os.path.exists(self.command_file))
        # There's still an update available though, so check again.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertFalse(downloading)
        # Now we'll let the download proceed to completion.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run(self.iface.DownloadUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # And now there is a command file for the update.
        self.assertTrue(os.path.exists(self.command_file))


class TestDBusGetSet(_TestBase):
    """Test the DBus client's key/value settings."""

    mode = 'live'

    def test_set_get_basic(self):
        # get/set a random key.
        self.iface.SetSetting('name', 'ant')
        self.assertEqual(self.iface.GetSetting('name'), 'ant')

    def test_set_get_change(self):
        # get/set a random key, then change it.
        self.iface.SetSetting('name', 'ant')
        self.assertEqual(self.iface.GetSetting('name'), 'ant')
        self.iface.SetSetting('name', 'bee')
        self.assertEqual(self.iface.GetSetting('name'), 'bee')

    def test_get_before_set(self):
        # Getting a key that doesn't exist returns the empty string.
        self.assertEqual(self.iface.GetSetting('thing'), '')
        self.iface.SetSetting('thing', 'one')
        self.assertEqual(self.iface.GetSetting('thing'), 'one')

    def test_setting_persists(self):
        # Set a key, restart the dbus server, and the key's value persists.
        self.iface.SetSetting('permanent', 'waves')
        self.assertEqual(self.iface.GetSetting('permanent'), 'waves')
        self.iface.Exit()
        self.assertRaises(DBusException, self.iface.GetSetting, 'permanent')
        # Re-establish a new connection.
        bus = dbus.SystemBus()
        service = bus.get_object('com.canonical.SystemImage', '/Service')
        self.iface = dbus.Interface(service, 'com.canonical.SystemImage')
        self.assertEqual(self.iface.GetSetting('permanent'), 'waves')

    def test_setting_min_battery_good(self):
        # min_battery has special semantics.
        self.iface.SetSetting('min_battery', '30')
        self.assertEqual(self.iface.GetSetting('min_battery'), '30')

    def test_setting_min_battery_bad(self):
        # min_battery requires the string representation of a percentage.
        self.iface.SetSetting('min_battery', 'standby')
        self.assertEqual(self.iface.GetSetting('min_battery'), '')
        self.iface.SetSetting('min_battery', '30')
        self.assertEqual(self.iface.GetSetting('min_battery'), '30')
        self.iface.SetSetting('min_battery', 'foo')
        self.assertEqual(self.iface.GetSetting('min_battery'), '30')
        self.iface.SetSetting('min_battery', '-10')
        self.assertEqual(self.iface.GetSetting('min_battery'), '30')
        self.iface.SetSetting('min_battery', '100')
        self.assertEqual(self.iface.GetSetting('min_battery'), '100')
        self.iface.SetSetting('min_battery', '101')
        self.assertEqual(self.iface.GetSetting('min_battery'), '100')
        self.iface.SetSetting('min_battery', 'standby')
        self.assertEqual(self.iface.GetSetting('min_battery'), '100')

    def test_setting_auto_download_good(self):
        # auto_download has special semantics.
        self.iface.SetSetting('auto_download', '0')
        self.assertEqual(self.iface.GetSetting('auto_download'), '0')
        self.iface.SetSetting('auto_download', '1')
        self.assertEqual(self.iface.GetSetting('auto_download'), '1')
        self.iface.SetSetting('auto_download', '2')
        self.assertEqual(self.iface.GetSetting('auto_download'), '2')

    def test_setting_auto_download_bad(self):
        # auto_download requires an integer between 0 and 2.  Don't forget
        # that it gets pre-populated when the database is created.
        self.iface.SetSetting('auto_download', 'standby')
        self.assertEqual(self.iface.GetSetting('auto_download'), '1')
        self.iface.SetSetting('auto_download', '-1')
        self.assertEqual(self.iface.GetSetting('auto_download'), '1')
        self.iface.SetSetting('auto_download', '0')
        self.assertEqual(self.iface.GetSetting('auto_download'), '0')
        self.iface.SetSetting('auto_download', '3')
        self.assertEqual(self.iface.GetSetting('auto_download'), '0')
        self.iface.SetSetting('auto_download', '2')
        self.assertEqual(self.iface.GetSetting('auto_download'), '2')

    def test_prepopulated_settings(self):
        # Some settings are pre-populated.
        self.assertEqual(self.iface.GetSetting('auto_download'), '1')

    def test_setting_changed_signal(self):
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'foo', 'yes'))
        self.assertEqual(len(reactor.signals), 1)
        key, new_value = reactor.signals[0]
        self.assertEqual(key, 'foo')
        self.assertEqual(new_value, 'yes')
        # The value did not change.
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'foo', 'yes'), timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        # This is the default value, so nothing changes.
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'auto_download', '1'),
                    timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        # This is a bogus value, so nothing changes.
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'min_battery', '200'),
                    timeout=15)
        self.assertEqual(len(reactor.signals), 0)
        # Change back.
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'auto_download', '0'))
        self.assertEqual(len(reactor.signals), 1)
        key, new_value = reactor.signals[0]
        self.assertEqual(key, 'auto_download')
        self.assertEqual(new_value, '0')
        # Change back.
        reactor = SignalCapturingReactor('SettingChanged')
        reactor.run(partial(self.iface.SetSetting, 'min_battery', '30'))
        self.assertEqual(len(reactor.signals), 1)
        key, new_value = reactor.signals[0]
        self.assertEqual(key, 'min_battery')
        self.assertEqual(new_value, '30')


class TestDBusInfo(_TestBase):
    mode = 'more-info'

    def test_info(self):
        # .Info() with a channel.ini containing version details.
        buildno, device, channel, last_update, details = self.iface.Info()
        self.assertEqual(buildno, 45)
        self.assertEqual(device, 'nexus11')
        self.assertEqual(channel, 'daily-proposed')
        self.assertEqual(last_update, '2099-08-01 04:45:45')
        self.assertEqual(details, dict(ubuntu='123', mako='456', custom='789'))


class TestDBusInfoNoDetails(_LiveTesting):
    def test_info_no_version_details(self):
        # .Info() where there is no channel.ini with version details.
        self._touch_build(45)
        timestamp = datetime(2022, 8, 1, 4, 45, 45).timestamp()
        os.utime(self.config.system.build_file, (timestamp, timestamp))
        buildno, device, channel, last_update, details = self.iface.Info()
        self.assertEqual(buildno, 45)
        self.assertEqual(device, 'nexus7')
        self.assertEqual(channel, 'stable')
        self.assertEqual(last_update, '2022-08-01 04:45:45')
        self.assertEqual(details, {})


class TestDBusProgress(_LiveTesting):
    def test_progress(self):
        self.download_always()
        self._touch_build(0)
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # Start the download and watch the progress meters.
        reactor = ProgressRecordingReactor()
        reactor.schedule(self.iface.DownloadUpdate)
        reactor.run()
        # The only progress we can count on is the first and last ones.  All
        # will have an eta of 0, since that value is not calculatable right
        # now.  The first progress will have percentage 0 and the last will
        # have percentage 100.
        self.assertGreaterEqual(len(reactor.progress), 2)
        percentage, eta = reactor.progress[0]
        self.assertEqual(percentage, 0)
        self.assertEqual(eta, 0)
        percentage, eta = reactor.progress[-1]
        self.assertEqual(percentage, 100)
        self.assertEqual(eta, 0)


class TestDBusPauseResume(_LiveTesting):
    def setUp(self):
        super().setUp()
        # We have to hack the files to be rather large so that the download
        # doesn't complete before we get a chance to pause it.  Of course,
        # this breaks the signatures because we're changing the file contents
        # after the .asc files have been written.
        for path in ('3/4/5.txt', '4/5/6.txt', '5/6/7.txt'):
            full_path = os.path.join(
                SystemImagePlugin.controller.serverdir, path)
            with open(full_path, 'wb') as fp:
                # 100 MiB
                for i in range(25600):
                    fp.write(b'x' * 4096)

    def test_pause(self):
        self.download_manually()
        self._touch_build(0)
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # We're ready to start downloading.  We schedule a pause to happen in
        # a little bit and then ensure that we get the proper signal.
        reactor = PausingReactor(self.iface)
        reactor.schedule(self.iface.DownloadUpdate)
        reactor.run(timeout=15)
        self.assertTrue(reactor.paused)
        # Now let's resume the download.  Because we intentionally corrupted
        # the downloaded files, we'll get an UpdateFailed signal instead of
        # the successful UpdateDownloaded signal.
        reactor = SignalCapturingReactor('UpdateFailed')
        reactor.run(self.iface.DownloadUpdate, timeout=60)
        self.assertEqual(len(reactor.signals), 1)
        # The error message will include lots of details on the SignatureError
        # that results.  The key thing is that it's 5.txt that is the first
        # file to fail its signature check.
        failure_count, last_error = reactor.signals[0]
        self.assertEqual(failure_count, 1)
        check_next = False
        for line in last_error.splitlines():
            line = line.strip()
            if check_next:
                self.assertEqual(os.path.basename(line), '5.txt')
                break
            if line.startswith('data path:'):
                check_next = True
        else:
            raise AssertionError('Did not find expected error output')


class TestDBusUseCache(_LiveTesting):
    # See LP: #1217098

    def test_use_cache(self):
        # We run the D-Bus service once through to download all the relevant
        # files.  Then we kill the service before performing the reboot, and
        # try to do another download.  The second one should only try to
        # download the ancillary files (i.e. channels.json, index.json,
        # keyrings), but not the data files.
        self.download_always()
        self._touch_build(0)
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        # There's one boolean argument to the result.
        (is_available, downloading, available_version, update_size,
         last_update_date,
         #descriptions,
         error_reason) = reactor.signals[0]
        self.assertTrue(is_available)
        self.assertTrue(downloading)
        # Now, wait for the UpdateDownloaded signal.
        reactor = SignalCapturingReactor('UpdateDownloaded')
        reactor.run()
        self.assertEqual(len(reactor.signals), 1)
        config = Configuration()
        config.load(SystemImagePlugin.controller.ini_path)
        self.assertEqual(set(os.listdir(config.updater.cache_partition)),
                         set((
                             '5.txt',
                             '5.txt.asc',
                             '6.txt',
                             '6.txt.asc',
                             '7.txt',
                             '7.txt.asc',
                             'device-signing.tar.xz',
                             'device-signing.tar.xz.asc',
                             'image-master.tar.xz',
                             'image-master.tar.xz.asc',
                             'image-signing.tar.xz',
                             'image-signing.tar.xz.asc',
                             'ubuntu_command',
                             )))
        # To prove that the data files are not downloaded again, let's
        # actually remove them from the server.
        for dirpath, dirnames, filenames in os.walk(
                SystemImagePlugin.controller.serverdir):
            for filename in filenames:
                if filename.endswith('.txt') or filename.endswith('.txt.asc'):
                    os.remove(os.path.join(dirpath, filename))
        # As extra proof, get the mtime in nanoseconds for the .txt and
        # .txt.asc files.
        mtimes = {}
        for filename in os.listdir(config.updater.cache_partition):
            path = os.path.join(config.updater.cache_partition, filename)
            if filename.endswith('.txt') or filename.endswith('.txt.asc'):
                mtimes[filename] = os.stat(path).st_mtime_ns
        # Don't issue the reboot.  Instead, kill the services, which throws
        # away all state, but does not delete the cached files.
        # Re-establish a new connection.
        process = find_dbus_process(SystemImagePlugin.controller.ini_path)
        if process is not None:
            self.iface.Exit()
            process.wait(60)
        bus = dbus.SystemBus()
        service = bus.get_object('com.canonical.SystemImage', '/Service')
        self.iface = dbus.Interface(service, 'com.canonical.SystemImage')
        # Now, if we just apply the update, it will succeed, since it knows
        # that the cached files are valid.
        reactor = SignalCapturingReactor('Rebooting')
        reactor.run(self.iface.ApplyUpdate)
        self.assertEqual(len(reactor.signals), 1)
        self.assertTrue(reactor.signals[0])
        self.assertEqual(set(os.listdir(config.updater.cache_partition)),
                         set((
                             '5.txt',
                             '5.txt.asc',
                             '6.txt',
                             '6.txt.asc',
                             '7.txt',
                             '7.txt.asc',
                             'device-signing.tar.xz',
                             'device-signing.tar.xz.asc',
                             'image-master.tar.xz',
                             'image-master.tar.xz.asc',
                             'image-signing.tar.xz',
                             'image-signing.tar.xz.asc',
                             # This file exists because reboot is mocked out
                             # in the dbus tests to just write to a log file.
                             'reboot.log',
                             'ubuntu_command',
                             )))
        for filename in os.listdir(config.updater.cache_partition):
            path = os.path.join(config.updater.cache_partition, filename)
            if filename.endswith('.txt') or filename.endswith('.txt.asc'):
                self.assertEqual(mtimes[filename], os.stat(path).st_mtime_ns)
        # Make sure the ubuntu_command file has the full update.
        path = os.path.join(config.updater.cache_partition, 'ubuntu_command')
        with open(path, 'r', encoding='utf-8') as fp:
            command = fp.read()
        self.assertMultiLineEqual(command, """\
load_keyring image-master.tar.xz image-master.tar.xz.asc
load_keyring image-signing.tar.xz image-signing.tar.xz.asc
load_keyring device-signing.tar.xz device-signing.tar.xz.asc
format system
mount system
update 6.txt 6.txt.asc
update 7.txt 7.txt.asc
update 5.txt 5.txt.asc
unmount system
""")


class TestDBusMultipleChecksInFlight(_LiveTesting):
    def test_multiple_check_for_updates(self):
        # Log analysis of LP: #1277589 appears to show the following scenario,
        # reproduced in this test case:
        #
        # * Automatic updates are enabled.
        # * No image signing or image master keys are present.
        # * A full update is checked.
        #   - A new image master key and image signing key is downloaded.
        #   - Update is available
        #
        # Start by creating some big files which will take a while to
        # download.
        def write_callback(dst):
            # Write a 100 MiB sized file.
            with open(dst, 'wb') as fp:
                for i in range(25600):
                    fp.write(b'x' * 4096)
        self._prepare_index('index_24.json', write_callback)
        # Create a reactor that will exit when the UpdateDownloaded signal is
        # received.  We're going to issue a CheckForUpdate with automatic
        # updates enabled.  As soon as we receive the UpdateAvailableStatus
        # signal, we'll immediately issue *another* CheckForUpdate, which
        # should run while the auto-download is working.
        #
        # As per LP: #1284217, we will get a second UpdateAvailableStatus
        # signal, since the status is available even while the original
        # request is being downloaded.
        reactor = DoubleCheckingReactor(self.iface)
        reactor.run()
        # We need to have received at least 2 signals, but due to timing
        # issues it could possibly be more.
        self.assertGreater(len(reactor.uas_signals), 1)
        # All received signals should have the same information.
        for signal in reactor.uas_signals:
            (is_available, downloading, available_version, update_size,
             last_update_date,
             #descriptions,
             error_reason) = signal
            self.assertTrue(is_available)
            self.assertTrue(downloading)
            self.assertEqual(available_version, '1600')
            self.assertEqual(update_size, 314572800)
            self.assertEqual(last_update_date, 'Unknown')
            self.assertEqual(error_reason, '')

    def test_multiple_check_for_updates_with_manual_downloading(self):
        # Log analysis of LP: #1287919 (a refinement of LP: #1277589 with
        # manual downloading enabled) shows that it's possible to enter the
        # checking phase while a download of the data files is still running.
        # When manually downloading, this will start another check, and as
        # part of that check, the blacklist and other files will be deleted
        # (in anticipation of them being re-downloaded).  When the data files
        # are downloaded, the state machine that just did the data download
        # may find its files deleted out from underneath it by the state
        # machine doing the checking.
        self.download_manually()
        # Start by creating some big files which will take a while to
        # download.
        def write_callback(dst):
            # Write a 100 MiB sized file.
            with open(dst, 'wb') as fp:
                for i in range(25600):
                    fp.write(b'x' * 4096)
        self._prepare_index('index_24.json', write_callback)
        self._touch_build(0)
        # Create a reactor that implements the following test plan:
        # * Set the device to download manually.
        # * Flash to an older revision
        # * Open System Settings and wait for it to say Updates available
        # * Click on About this phone
        # * Click on Check for Update and wait for it to say Install 1 update
        # * Click on Install 1 update and while the files are downloading,
        #   swipe up from the bottom and click Back
        # * Click on Check for Update again
        # * Wait for the Update System overlay to come up, and then install
        #   the update, and reboot
        reactor = ManualUpdateReactor(self.iface)
        reactor.run()
        self.assertTrue(reactor.rebooting)


class TestDBusCheckForUpdateException(_LiveTesting):
    @classmethod
    def setUpClass(cls):
        ini_path = SystemImagePlugin.controller.ini_path
        with open(ini_path, 'r', encoding='utf-8') as fp:
            lines = fp.readlines()
        with atomic(ini_path) as fp:
            for line in lines:
                key, sep, value = line.partition(': ')
                if key == 'cache_partition':
                    cls.bad_path = os.path.join(value.strip(), 'unwritable')
                    print('{}: {}'.format(key, cls.bad_path), file=fp)
                    os.makedirs(cls.bad_path, mode=0)
                else:
                    fp.write(line)
        super().setUpClass()

    def tearDown(self):
        os.chmod(self.bad_path, 0o777)
        super().tearDown()

    def test_check_for_update_error(self):
        # CheckForUpdate sees an error, in this case because the destination
        # directory for downloads is not writable.  We'll get an
        # UpdateAvailableStatus with an error string.
        reactor = SignalCapturingReactor('UpdateAvailableStatus')
        reactor.run(self.iface.CheckForUpdate)
        self.assertEqual(len(reactor.signals), 1)
        (is_available, downloading, available_version, update_size,
         last_update_date,
         # descriptions,
         error_reason) = reactor.signals[0]
        self.assertIn('Permission denied', error_reason)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""DBus service main entry point."""

__all__ = [
    'main',
    ]


import os
import sys
import dbus
import logging
import argparse

from contextlib import ExitStack
from dbus.mainloop.glib import DBusGMainLoop
from pkg_resources import resource_string as resource_bytes
from systemimage.config import config
from systemimage.dbus import Loop
from systemimage.helpers import makedirs
from systemimage.logging import initialize
from systemimage.main import DEFAULT_CONFIG_FILE

# --testing is only enabled when the systemimage.testing package is
# available.  This will be the case for the upstream source package, and when
# the systemimage-dev binary package is installed in Ubuntu.
try:
    from systemimage.testing.dbus import instrument, get_service
except ImportError:
    instrument = None
    get_service = None


__version__ = resource_bytes(
    'systemimage', 'version.txt').decode('utf-8').strip()


def main():
    global config
    parser = argparse.ArgumentParser(
        prog='system-image-dbus',
        description='Ubuntu System Image Upgrader DBus service')
    parser.add_argument('--version',
                        action='version',
                        version='system-image-dbus {}'.format(__version__))
    parser.add_argument('-C', '--config',
                        default=DEFAULT_CONFIG_FILE, action='store',
                        metavar='FILE',
                        help="""Use the given configuration file instead of
                                the default""")
    parser.add_argument('-v', '--verbose',
                        default=0, action='count',
                        help='Increase verbosity')
    # Hidden argument for special setup required by test environment.
    if instrument is not None:
        parser.add_argument('--testing',
                            default=False, action='store',
                            help=argparse.SUPPRESS)

    args = parser.parse_args(sys.argv[1:])
    try:
        config.load(args.config)
    except FileNotFoundError as error:
        parser.error('\nConfiguration file not found: {}'.format(error))
        assert 'parser.error() does not return'
    # Load the optional channel.ini file, which must live next to the
    # configuration file.  It's okay if this file does not exist.
    channel_ini = os.path.join(os.path.dirname(args.config), 'channel.ini')
    try:
        config.load(channel_ini, override=True)
    except FileNotFoundError:
        pass

    # Create the temporary directory if it doesn't exist.
    makedirs(config.system.tempdir)
    # Initialize the loggers.
    initialize(verbosity=args.verbose)
    log = logging.getLogger('systemimage')

    DBusGMainLoop(set_as_default=True)

    system_bus = dbus.SystemBus()
    # Ensure we're the only owner of this bus name.
    code = system_bus.request_name(
        'com.canonical.SystemImage',
        dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
    if code == dbus.bus.REQUEST_NAME_REPLY_EXISTS:
        # Another instance already owns this name.  Exit.
        log.error('Cannot get exclusive ownership of bus name.')
        sys.exit(2)

    log.info('SystemImage dbus main loop starting [{}/{}]',
             config.channel, config.device)

    with ExitStack() as stack:
        loop = Loop()
        testing_mode = getattr(args, 'testing', None)
        if testing_mode:
            instrument(config, stack)
            config.dbus_service = get_service(
                testing_mode, system_bus, '/Service', loop)
        else:
            from systemimage.dbus import Service
            config.dbus_service = Service(system_bus, '/Service', loop)
        try:
            loop.run()
        except KeyboardInterrupt:
            log.info('SystemImage dbus main loop interrupted')
        except:
            log.exception('D-Bus loop exception')
            raise
        else:
            log.info('SystemImage dbus main loop exited')


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Device/channel indexes."""

__all__ = [
    'Index',
    ]


import json

from datetime import datetime, timezone
from systemimage.bag import Bag
from systemimage.helpers import phased_percentage
from systemimage.image import Image


IN_FMT = '%a %b %d %H:%M:%S %Z %Y'
OUT_FMT = '%a %b %d %H:%M:%S UTC %Y'


class Index(Bag):
    @classmethod
    def from_json(cls, data):
        """Parse the JSON data and produce an index."""
        mapping = json.loads(data)
        # Parse the global data, of which there is only the timestamp.  Even
        # though the string will contain 'UTC' (which we assert is so since we
        # can only handle UTC timestamps), strptime() will return a naive
        # datetime.  We'll turn it into an aware datetime in UTC, which is the
        # only thing that can possibly make sense.
        timestamp_str = mapping['global']['generated_at']
        assert 'UTC' in timestamp_str.split(), 'timestamps must be UTC'
        naive_generated_at = datetime.strptime(timestamp_str, IN_FMT)
        generated_at=naive_generated_at.replace(tzinfo=timezone.utc)
        global_ = Bag(generated_at=generated_at)
        # Parse the images.
        images = []
        percentage = phased_percentage()
        for image_data in mapping['images']:
            # Descriptions can be any of:
            #
            # * description
            # * description-xx (e.g. description-en)
            # * description-xx_CC (e.g. description-en_US)
            #
            # We want to preserve the keys exactly as given, and because the
            # extended forms are not Python identifiers, we'll pull these out
            # into a separate, non-Bag dictionary.
            descriptions = {}
            # We're going to mutate the dictionary during iteration.
            for key in list(image_data):
                if key.startswith('description'):
                    descriptions[key] = image_data.pop(key)
            files = image_data.pop('files', [])
            bundles = [Bag(**bundle_data) for bundle_data in files]
            image = Image(files=bundles,
                          descriptions=descriptions,
                          **image_data)
            if percentage <= image.phased_percentage:
                images.append(image)
        return cls(global_=global_, images=images)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the candidate upgrade path algorithm."""

__all__ = [
    'TestCandidateDownloads',
    'TestCandidateFilters',
    'TestCandidates',
    'TestNewVersionRegime',
    ]


import unittest

from operator import attrgetter
from systemimage.candidates import (
    delta_filter, full_filter, get_candidates, iter_path)
from systemimage.scores import WeightedScorer
from systemimage.testing.helpers import configuration, get_index


def _descriptions(path):
    descriptions = []
    for image in path:
        # There's only one description per image so order doesn't
        # matter.
        descriptions.extend(image.descriptions.values())
    return descriptions


class TestCandidates(unittest.TestCase):
    def test_no_images(self):
        # If there are no images defined, there are no candidates.
        index = get_index('index_01.json')
        candidates = get_candidates(index, 1400)
        self.assertEqual(candidates, [])

    def test_only_higher_fulls(self):
        # All the full images have a minversion greater than our version, so
        # we cannot upgrade to any of them.
        index = get_index('index_02.json')
        candidates = get_candidates(index, 100)
        self.assertEqual(candidates, [])

    def test_one_higher_full(self):
        # Our device is between the minversions of the two available fulls, so
        # the older one can be upgraded too.
        index = get_index('index_02.json')
        candidates = get_candidates(index, 800)
        # There is exactly one upgrade path.
        self.assertEqual(len(candidates), 1)
        path = candidates[0]
        # The path has exactly one image.
        self.assertEqual(len(path), 1)
        image = path[0]
        self.assertEqual(list(image.descriptions.values()),
                         ['New full build 1'])

    def test_fulls_with_no_minversion(self):
        # Like the previous test, there are two full upgrades, but because
        # neither of them have minversions, both are candidates.
        index = get_index('index_05.json')
        candidates = get_candidates(index, 400)
        self.assertEqual(len(candidates), 2)
        # Both candidate paths have exactly one image in them.  We can't sort
        # these paths, so just test them both.
        path0, path1 = candidates
        self.assertEqual(len(path0), 1)
        self.assertEqual(len(path1), 1)
        # One path gets us to version 1300 and the other 1400.
        images = sorted([path0[0], path1[0]], key=attrgetter('version'))
        self.assertEqual(list(images[0].descriptions.values()),
                         ['New full build 1'])
        self.assertEqual(list(images[1].descriptions.values()),
                         ['New full build 2'])

    def test_no_deltas_based_on_us(self):
        # There are deltas in the test data, but no fulls.  None of the deltas
        # have a base equal to our build number.
        index = get_index('index_03.json')
        candidates = get_candidates(index, 100)
        self.assertEqual(candidates, [])

    def test_one_delta_based_on_us(self):
        # There is one delta in the test data that is based on us.
        index = get_index('index_03.json')
        candidates = get_candidates(index, 500)
        self.assertEqual(len(candidates), 1)
        path = candidates[0]
        # The path has exactly one image.
        self.assertEqual(len(path), 1)
        image = path[0]
        self.assertEqual(list(image.descriptions.values()), ['Delta 2'])

    def test_two_deltas_based_on_us(self):
        # There are two deltas that are based on us, so both are candidates.
        # They get us to different final versions.
        index = get_index('index_04.json')
        candidates = get_candidates(index, 1100)
        self.assertEqual(len(candidates), 2)
        # Both candidate paths have exactly one image in them.  We can't sort
        # these paths, so just test them both.
        path0, path1 = candidates
        self.assertEqual(len(path0), 1)
        self.assertEqual(len(path1), 1)
        # One path gets us to version 1300 and the other 1400.
        images = sorted([path0[0], path1[0]], key=attrgetter('version'))
        self.assertEqual(_descriptions(images), ['Delta 2', 'Delta 1'])

    def test_one_path_with_full_and_deltas(self):
        # There's one path to upgrade from our version to the final version.
        # This one starts at a full and includes several deltas.
        index = get_index('index_06.json')
        candidates = get_candidates(index, 1000)
        self.assertEqual(len(candidates), 1)
        path = candidates[0]
        self.assertEqual(len(path), 3)
        self.assertEqual([image.version for image in path],
                         [1300, 1301, 1302])
        self.assertEqual(_descriptions(path), ['Full 1', 'Delta 1', 'Delta 2'])

    def test_one_path_with_deltas(self):
        # Similar to above, except that because we're upgrading from the
        # version of the full, the path is only two images long, i.e. the
        # deltas.
        index = get_index('index_06.json')
        candidates = get_candidates(index, 1300)
        self.assertEqual(len(candidates), 1)
        path = candidates[0]
        self.assertEqual(len(path), 2)
        self.assertEqual([image.version for image in path], [1301, 1302])
        self.assertEqual(_descriptions(path), ['Delta 1', 'Delta 2'])

    def test_forked_paths(self):
        # We have a fork in the road.  There is a full update, but two deltas
        # with different versions point to the same base.  This will give us
        # two upgrade paths, both of which include the full.
        index = get_index('index_07.json')
        candidates = get_candidates(index, 1200)
        self.assertEqual(len(candidates), 2)
        # We can sort the paths by length.
        paths = sorted(candidates, key=len)
        # The shortest path gets us to 1302 in two steps.
        self.assertEqual(len(paths[0]), 2)
        self.assertEqual([image.version for image in paths[0]], [1300, 1302])
        descriptions = []
        for image in paths[0]:
            # There's only one description per image so order doesn't matter.
            descriptions.extend(image.descriptions.values())
        self.assertEqual(descriptions, ['Full 1', 'Delta 2'])
        # The longer path gets us to 1302 in three steps.
        self.assertEqual(len(paths[1]), 3)
        self.assertEqual([image.version for image in paths[1]],
                         [1300, 1301, 1302])
        descriptions = []
        for image in paths[1]:
            # There's only one description per image so order doesn't matter.
            descriptions.extend(image.descriptions.values())
        self.assertEqual(descriptions, ['Full 1', 'Delta 1', 'Delta 3'])


class TestCandidateDownloads(unittest.TestCase):
    maxDiff = None

    @configuration
    def test_get_downloads(self):
        # Path B will win; it has one full and two deltas, none of which have
        # a bootme flag.  Download all their files.
        index = get_index('index_10.json')
        candidates = get_candidates(index, 600)
        winner = WeightedScorer().choose(candidates)
        descriptions = []
        for image in winner:
            # There's only one description per image so order doesn't matter.
            descriptions.extend(image.descriptions.values())
        self.assertEqual(descriptions, ['Full B', 'Delta B.1', 'Delta B.2'])
        downloads = list(iter_path(winner))
        paths = set(filerec.path for (n, filerec) in downloads)
        self.assertEqual(paths, set([
            '/3/4/5.txt',
            '/4/5/6.txt',
            '/5/6/7.txt',
            '/6/7/8.txt',
            '/7/8/9.txt',
            '/8/9/a.txt',
            '/9/a/b.txt',
            '/e/d/c.txt',
            '/f/e/d.txt',
            ]))
        signatures = set(filerec.signature for (n, filerec) in downloads)
        self.assertEqual(signatures, set([
            '/3/4/5.txt.asc',
            '/4/5/6.txt.asc',
            '/5/6/7.txt.asc',
            '/6/7/8.txt.asc',
            '/7/8/9.txt.asc',
            '/8/9/a.txt.asc',
            '/9/a/b.txt.asc',
            '/e/d/c.txt.asc',
            '/f/e/d.txt.asc',
            ]))

    @configuration
    def test_get_downloads_with_bootme(self):
        # Path B will win; it has one full and two deltas.  The first delta
        # has a bootme flag so the second delta's files are not downloaded.
        index = get_index('index_11.json')
        candidates = get_candidates(index, 600)
        winner = WeightedScorer().choose(candidates)
        descriptions = []
        for image in winner:
            # There's only one description per image so order doesn't matter.
            descriptions.extend(image.descriptions.values())
        self.assertEqual(descriptions, ['Full B', 'Delta B.1', 'Delta B.2'])
        downloads = iter_path(winner)
        paths = set(filerec.path for (n, filerec) in downloads)
        self.assertEqual(paths, set([
            '/3/4/5.txt',
            '/4/5/6.txt',
            '/5/6/7.txt',
            '/6/7/8.txt',
            '/7/8/9.txt',
            '/8/9/a.txt',
            ]))


class TestCandidateFilters(unittest.TestCase):
    def test_filter_for_fulls(self):
        # Run a filter over the candidates, such that the only ones left are
        # those that contain only full upgrades.  This can truncate any paths
        # that start with some fulls and then contain some deltas.
        index = get_index('index_10.json')
        candidates = get_candidates(index, 600)
        filtered = full_filter(candidates)
        # Since all images start with a full update, we're still left with
        # three candidates.
        self.assertEqual(len(filtered), 3)
        self.assertEqual([image.type for image in filtered[0]], ['full'])
        self.assertEqual([image.type for image in filtered[1]], ['full'])
        self.assertEqual([image.type for image in filtered[2]], ['full'])
        self.assertEqual(_descriptions(filtered[0]), ['Full A'])
        self.assertEqual(_descriptions(filtered[1]), ['Full B'])
        self.assertEqual(_descriptions(filtered[2]), ['Full C'])

    def test_filter_for_fulls_one_candidate(self):
        # Filter for full updates, where the only candidate has one full image.
        index = get_index('index_13.json')
        candidates = get_candidates(index, 600)
        filtered = full_filter(candidates)
        self.assertEqual(filtered, candidates)

    def test_filter_for_fulls_with_just_delta_candidates(self):
        # A candidate path that contains only deltas will have no filtered
        # paths if all the images are delta updates.
        index = get_index('index_15.json')
        candidates = get_candidates(index, 100)
        self.assertEqual(len(candidates), 1)
        filtered = full_filter(candidates)
        self.assertEqual(len(filtered), 0)

    def test_filter_for_deltas(self):
        # Filter the candidates, where the only available path is a delta path.
        index = get_index('index_15.json')
        candidates = get_candidates(index, 100)
        self.assertEqual(len(candidates), 1)
        filtered = delta_filter(candidates)
        self.assertEqual(len(filtered), 1)
        self.assertEqual(candidates, filtered)

    def test_filter_for_deltas_none_available(self):
        # Run a filter over the candidates, such that the only ones left are
        # those that start with and contain only deltas.  Since none of the
        # paths do so, tere are no candidates left.
        index = get_index('index_10.json')
        candidates = get_candidates(index, 600)
        filtered = delta_filter(candidates)
        self.assertEqual(len(filtered), 0)

    def test_filter_for_deltas_one_candidate(self):
        # Filter for delta updates, but the only candidate is a full.
        index = get_index('index_13.json')
        candidates = get_candidates(index, 600)
        filtered = delta_filter(candidates)
        self.assertEqual(len(filtered), 0)

    def test_filter_for_multiple_deltas(self):
        # The candidate path has multiple deltas.  All are preserved.
        index = get_index('index_19.json')
        candidates = get_candidates(index, 100)
        filtered = delta_filter(candidates)
        self.assertEqual(len(filtered), 1)
        path = filtered[0]
        self.assertEqual(len(path), 3)
        self.assertEqual(_descriptions(path),
                         ['Delta A', 'Delta B', 'Delta C'])


class TestNewVersionRegime(unittest.TestCase):
    """LP: #1218612"""

    def test_candidates(self):
        # Path B will win; it has one full and two deltas.
        index = get_index('index_20.json')
        candidates = get_candidates(index, 0)
        self.assertEqual(len(candidates), 3)
        path0 = candidates[0]
        self.assertEqual(_descriptions(path0),
                         ['Full A', 'Delta A.1', 'Delta A.2'])
        path1 = candidates[1]
        self.assertEqual(_descriptions(path1),
                         ['Full B', 'Delta B.1', 'Delta B.2'])
        path2 = candidates[2]
        self.assertEqual(_descriptions(path2), ['Full C', 'Delta C.1'])
        # The version numbers use the new regime.
        self.assertEqual(path0[0].version, 300)
        self.assertEqual(path0[1].base, 300)
        self.assertEqual(path0[1].version, 301)
        self.assertEqual(path0[2].base, 301)
        self.assertEqual(path0[2].version, 304)
        winner = WeightedScorer().choose(candidates)
        self.assertEqual(_descriptions(winner),
                         ['Full B', 'Delta B.1', 'Delta B.2'])
        self.assertEqual(winner[0].version, 200)
        self.assertEqual(winner[1].base, 200)
        self.assertEqual(winner[1].version, 201)
        self.assertEqual(winner[2].base, 201)
        self.assertEqual(winner[2].version, 304)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test channel/device index parsing."""

__all__ = [
    'TestIndex',
    'TestDownloadIndex',
    ]


import os
import unittest

from contextlib import ExitStack
from datetime import datetime, timezone
from systemimage.gpg import SignatureError
from systemimage.helpers import temporary_directory
from systemimage.state import State
from systemimage.testing.helpers import (
    configuration, copy, get_index, make_http_server, makedirs,
    setup_keyring_txz, setup_keyrings, sign)
from systemimage.testing.nose import SystemImagePlugin
# FIXME
from systemimage.tests.test_candidates import _descriptions
from unittest.mock import patch


class TestIndex(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        SystemImagePlugin.controller.set_mode(cert_pem='cert.pem')

    def test_index_global(self):
        index = get_index('index_01.json')
        self.assertEqual(
            index.global_.generated_at,
            datetime(2013, 4, 29, 18, 45, 27, tzinfo=timezone.utc))

    def test_index_image_count(self):
        index = get_index('index_01.json')
        self.assertEqual(len(index.images), 0)
        index = get_index('index_02.json')
        self.assertEqual(len(index.images), 2)

    def test_image_20130300_full(self):
        index = get_index('sprint_nexus7_index_01.json')
        image = index.images[0]
        self.assertEqual(
            image.descriptions,
            {'description': 'Some kind of daily build'})
        self.assertEqual(image.type, 'full')
        self.assertEqual(image.version, 20130300)
        self.assertTrue(image.bootme)
        self.assertEqual(len(image.files), 3)
        # The first file is the device dependent image.  The second is the
        # device independent file, and the third is the version zip.
        dev, ind, ver = image.files
        self.assertEqual(dev.path, '/sprint/nexus7/nexus7-20130300.full.zip')
        self.assertEqual(dev.signature,
                         '/sprint/nexus7/nexus7-20130300.full.zip.asc')
        self.assertEqual(dev.checksum, 'abcdef0')
        self.assertEqual(dev.order, 0)
        self.assertEqual(dev.size, 0)
        # Let's not check the whole file, just a few useful bits.
        self.assertEqual(ind.checksum, 'abcdef1')
        self.assertEqual(ind.order, 0)
        self.assertEqual(ver.checksum, 'abcdef2')
        self.assertEqual(ver.order, 1)

    def test_image_20130500_minversion(self):
        # Some full images have a minimum version older than which they refuse
        # to upgrade from.
        index = get_index('sprint_nexus7_index_01.json')
        image = index.images[5]
        self.assertEqual(image.type, 'full')
        self.assertEqual(image.version, 20130500)
        self.assertTrue(image.bootme)
        self.assertEqual(image.minversion, 20130100)

    def test_image_descriptions(self):
        # Image descriptions can come in a variety of locales.
        index = get_index('index_14.json')
        self.assertEqual(index.images[0].descriptions, {
            'description': 'Full A'})
        self.assertEqual(index.images[3].descriptions, {
            'description': 'Full B',
            'description-en': 'The full B',
            })
        self.assertEqual(index.images[4].descriptions, {
            'description': 'Delta B.1',
            'description-en_US': 'This is the delta B.1',
            'description-xx': 'XX This is the delta B.1',
            'description-yy': 'YY This is the delta B.1',
            'description-yy_ZZ': 'YY-ZZ This is the delta B.1',
            })
        # The second delta.
        self.assertEqual(index.images[5].descriptions, {
            'description': 'Delta B.2',
            'description-xx': 'Oh delta, my delta',
            'description-xx_CC': 'This hyar is the delta B.2',
            })

    def test_image_phased_percentage(self):
        # This index has two full updates with a phased-percentage value and
        # one without (which defaults to 100).  We'll set the system's
        # percentage right in the middle of the two so that the one with 50%
        # will not show up in the list of images.
        with patch('systemimage.index.phased_percentage', return_value=66):
            index = get_index('index_22.json')
        descriptions = set(_descriptions(index.images))
        # This one does not have a phased-percentage, so using the default of
        # 100, it gets in.
        self.assertIn('Full A', descriptions)
        # This one has a phased-percentage of 50 so it gets ignored.
        self.assertNotIn('Full B', descriptions)
        # This one has a phased-percentage of 75 so it gets added.
        self.assertIn('Full C', descriptions)

    def test_image_phased_percentage_100(self):
        # Like above, but with a system percentage of 100, so nothing but the
        # default gets in.
        with patch('systemimage.index.phased_percentage', return_value=100):
            index = get_index('index_22.json')
        descriptions = set(_descriptions(index.images))
        # This one does not have a phased-percentage, so using the default of
        # 100, it gets in.
        self.assertIn('Full A', descriptions)
        # This one has a phased-percentage of 50 so it gets ignored.
        self.assertNotIn('Full B', descriptions)
        # This one has a phased-percentage of 75 so it gets added.
        self.assertNotIn('Full C', descriptions)

    def test_image_phased_percentage_0(self):
        # Like above, but with a system percentage of 0, everything gets in.
        with patch('systemimage.index.phased_percentage', return_value=0):
            index = get_index('index_22.json')
        descriptions = set(_descriptions(index.images))
        self.assertIn('Full A', descriptions)
        self.assertIn('Full B', descriptions)
        self.assertIn('Full C', descriptions)


class TestDownloadIndex(unittest.TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(self):
        SystemImagePlugin.controller.set_mode(cert_pem='cert.pem')

    def setUp(self):
        # Start the HTTPS server running.  Vend it out of a temporary
        # directory which we load up with the right files.
        self._stack = ExitStack()
        try:
            self._serverdir = self._stack.enter_context(temporary_directory())
            self._stack.push(make_http_server(
                self._serverdir, 8943, 'cert.pem', 'key.pem'))
        except:
            self._stack.close()
            raise

    def tearDown(self):
        self._stack.close()

    def _copysign(self, src, dst, keyring):
        server_dst = os.path.join(self._serverdir, dst)
        makedirs(os.path.dirname(server_dst))
        copy(src, self._serverdir, dst)
        sign(server_dst, keyring)

    @configuration
    def test_load_index_good_path(self):
        # Load the index.json pointed to by the channels.json.  All signatures
        # validate correctly and there is no device keyring or blacklist.
        self._copysign(
            'channels_02.json', 'channels.json', 'image-signing.gpg')
        # index_10.json path B will win, with no bootme flags.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'image-signing.gpg')
        setup_keyrings()
        state = State()
        state.run_thru('get_index')
        self.assertEqual(
            state.index.global_.generated_at,
            datetime(2013, 4, 29, 18, 45, 27, tzinfo=timezone.utc))
        self.assertEqual(
            state.index.images[0].files[1].checksum, 'bcd')

    @configuration
    def test_load_index_with_device_keyring(self):
        # Here, the index.json file is signed with a device keyring.
        self._copysign(
            'channels_03.json', 'channels.json', 'image-signing.gpg')
        # index_10.json path B will win, with no bootme flags.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'device-signing.gpg')
        setup_keyrings()
        setup_keyring_txz(
            'device-signing.gpg', 'image-signing.gpg',
            dict(type='device-signing'),
            os.path.join(self._serverdir, 'stable', 'nexus7', 'device.tar.xz'))
        state = State()
        state.run_thru('get_index')
        self.assertEqual(
            state.index.global_.generated_at,
            datetime(2013, 4, 29, 18, 45, 27, tzinfo=timezone.utc))
        self.assertEqual(
            state.index.images[0].files[1].checksum, 'bcd')

    @configuration
    def test_load_index_with_device_keyring_and_signing_key(self):
        # Here, the index.json file is signed with the image signing keyring,
        # even though there is a device key.  That's fine.
        self._copysign(
            'channels_03.json', 'channels.json', 'image-signing.gpg')
        # index_10.json path B will win, with no bootme flags.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'image-signing.gpg')
        setup_keyrings()
        setup_keyring_txz(
            'device-signing.gpg', 'image-signing.gpg',
            dict(type='device-signing'),
            os.path.join(self._serverdir, 'stable', 'nexus7', 'device.tar.xz'))
        state = State()
        state.run_thru('get_index')
        self.assertEqual(
            state.index.global_.generated_at,
            datetime(2013, 4, 29, 18, 45, 27, tzinfo=timezone.utc))
        self.assertEqual(
            state.index.images[0].files[1].checksum, 'bcd')

    @configuration
    def test_load_index_with_bad_keyring(self):
        # Here, the index.json file is signed with a defective device keyring.
        self._copysign(
            'channels_03.json', 'channels.json', 'image-signing.gpg')
        # This will be signed by a keyring that is not the device keyring.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'spare.gpg')
        setup_keyrings()
        setup_keyring_txz(
            'device-signing.gpg', 'image-signing.gpg',
            dict(type='device-signing'),
            os.path.join(self._serverdir, 'stable', 'nexus7', 'device.tar.xz'))
        state = State()
        state.run_until('get_index')
        self.assertRaises(SignatureError, next, state)

    @configuration
    def test_load_index_with_blacklist(self):
        # Here, we've blacklisted the device key.
        self._copysign(
            'channels_03.json', 'channels.json', 'image-signing.gpg')
        # This will be signed by a keyring that is not the device keyring.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'device-signing.gpg')
        setup_keyrings()
        setup_keyring_txz(
            'device-signing.gpg', 'image-signing.gpg',
            dict(type='device-signing'),
            os.path.join(self._serverdir, 'stable', 'nexus7', 'device.tar.xz'))
        setup_keyring_txz(
            'device-signing.gpg', 'image-master.gpg', dict(type='blacklist'),
            os.path.join(self._serverdir, 'gpg', 'blacklist.tar.xz'))
        state = State()
        state.run_until('get_index')
        self.assertRaises(SignatureError, next, state)

    @configuration
    def test_missing_channel(self):
        # The system's channel does not exist.
        self._copysign(
            'channels_04.json', 'channels.json', 'image-signing.gpg')
        # index_10.json path B will win, with no bootme flags.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'image-signing.gpg')
        setup_keyrings()
        # Our channel (stable) isn't in the channels.json file, so there's
        # nothing to do.  Running the state machine to its conclusion leaves
        # us with no index file.
        state = State()
        list(state)
        # There really is nothing left to do.
        self.assertIsNone(state.index)

    @configuration
    def test_missing_device(self):
        # The system's device does not exist.
        self._copysign(
            'channels_05.json', 'channels.json', 'image-signing.gpg')
        # index_10.json path B will win, with no bootme flags.
        self._copysign(
            'index_10.json', 'stable/nexus7/index.json', 'image-signing.gpg')
        setup_keyrings()
        # Our device (nexus7) isn't in the channels.json file, so there's
        # nothing to do.  Running the state machine to its conclusion leaves
        # us with no index file.
        state = State()
        list(state)
        # There really is nothing left to do.
        self.assertIsNone(state.index)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Main script entry point."""


__all__ = [
    'main',
    ]


import os
import sys
import logging
import argparse

from dbus.mainloop.glib import DBusGMainLoop
from pkg_resources import resource_string as resource_bytes
from systemimage.bindings import DBusClient
from systemimage.candidates import delta_filter, full_filter
from systemimage.config import config
from systemimage.helpers import last_update_date, makedirs, version_detail
from systemimage.logging import initialize
from systemimage.state import State
from textwrap import dedent


__version__ = resource_bytes(
    'systemimage', 'version.txt').decode('utf-8').strip()

DEFAULT_CONFIG_FILE = '/etc/system-image/client.ini'
COLON = ':'


def main():
    global config
    parser = argparse.ArgumentParser(
        prog='system-image-cli',
        description='Ubuntu System Image Upgrader')
    parser.add_argument('--version',
                        action='version',
                        version='system-image-cli {}'.format(__version__))
    parser.add_argument('-C', '--config',
                        default=DEFAULT_CONFIG_FILE, action='store',
                        metavar='FILE',
                        help="""Use the given configuration file instead of
                                the default""")
    parser.add_argument('-b', '--build',
                        default=None, action='store',
                        help="""Override the current build number just
                                this once""")
    parser.add_argument('-c', '--channel',
                        default=None, action='store',
                        help="""Override the channel just this once.  Use in
                                combination with `--build 0` to switch
                                channels.""")
    parser.add_argument('-d', '--device',
                        default=None, action='store',
                        help='Override the device name just this once')
    parser.add_argument('--dbus',
                        default=False, action='store_true',
                        help='Run in D-Bus client mode.')
    parser.add_argument('-f', '--filter',
                        default=None, action='store',
                        help="""Filter the candidate paths to contain only
                                full updates or only delta updates.  The
                                argument to this option must be either `full`
                                or `delta`""")
    parser.add_argument('-i', '--info',
                        default=False, action='store_true',
                        help="""Show some information about the current
                                device, including the current build number,
                                device name and channel, then exit""")
    parser.add_argument('-n', '--dry-run',
                        default=False, action='store_true',
                        help="""Calculate and print the upgrade path, but do
                                not download or apply it""")
    parser.add_argument('-v', '--verbose',
                        default=0, action='count',
                        help='Increase verbosity')

    args = parser.parse_args(sys.argv[1:])
    try:
        config.load(args.config)
    except FileNotFoundError as error:
        parser.error('\nConfiguration file not found: {}'.format(error))
        assert 'parser.error() does not return'
    # Load the optional channel.ini file, which must live next to the
    # configuration file.  It's okay if this file does not exist.
    channel_ini = os.path.join(os.path.dirname(args.config), 'channel.ini')
    try:
        config.load(channel_ini, override=True)
    except FileNotFoundError:
        pass

    # Sanity check -f/--filter.
    if args.filter is None:
        candidate_filter = None
    elif args.filter == 'full':
        candidate_filter = full_filter
    elif args.filter == 'delta':
        candidate_filter = delta_filter
    else:
        parser.error('Bad filter type: {}'.format(args.filter))
        assert 'parser.error() does not return'

    # Create the temporary directory if it doesn't exist.
    makedirs(config.system.tempdir)
    # Initialize the loggers.
    initialize(verbosity=args.verbose)
    log = logging.getLogger('systemimage')
    # We assume the cache_partition already exists, as does the /etc directory
    # (i.e. where the archive master key lives).

    # Command line overrides.
    if args.build is not None:
        try:
            config.build_number = int(args.build)
        except ValueError:
            parser.error(
                '-b/--build requires an integer: {}'.format(args.build))
            assert 'parser.error() does not return'
    if args.channel is not None:
        config.channel = args.channel
    if args.device is not None:
        config.device = args.device

    if args.info:
        alias = getattr(config.service, 'channel_target', None)
        kws = dict(
            build_number=config.build_number,
            device=config.device,
            channel=config.channel,
            last_update=last_update_date(),
            )
        if alias is None:
            template = """\
                current build number: {build_number}
                device name: {device}
                channel: {channel}
                last update: {last_update}"""
        else:
            template = """\
                current build number: {build_number}
                device name: {device}
                channel: {channel}
                alias: {alias}
                last update: {last_update}"""
            kws['alias'] = alias
        print(dedent(template).format(**kws))
        # If there's additional version details, print this out now too.  We
        # sort the keys in reverse order because we want 'ubuntu' to generally
        # come first.
        details = version_detail()
        for key in sorted(details, reverse=True):
            print('version {}: {}'.format(key, details[key]))
        return 0

    # We can either run the API directly or through DBus.
    if args.dbus:
        client = DBusClient()
        client.check_for_update()
        if not client.is_available:
            log.info('No update is available')
            return 0
        if not client.downloaded:
            log.info('No update was downloaded')
            return 1
        if client.failed:
            log.info('Update failed')
            return 2
        client.reboot()
        # We probably won't get here..
        return 0

    # When verbosity is at 1, logging every progress signal from
    # ubuntu-download-manager would be way too noisy.  OTOH, not printing
    # anything leads some folks to think the process is just hung, since it
    # can take a long time to download all the data files.  As a compromise,
    # we'll output some dots to stderr at verbosity 1, but we won't log these
    # dots since they would just be noise.  This doesn't have to be perfect.
    if args.verbose == 1:
        dot_count = 0
        def callback(received, total):
            nonlocal dot_count
            sys.stderr.write('.')
            sys.stderr.flush()
            dot_count += 1
            if dot_count % 78 == 0 or received >= total:
                sys.stderr.write('\n')
                sys.stderr.flush()
    else:
        def callback(received, total):
            log.debug('received: {} of {} bytes', received, total)

    DBusGMainLoop(set_as_default=True)
    state = State(candidate_filter=candidate_filter)
    state.downloader.callback = callback
    if args.dry_run:
        try:
            state.run_until('download_files')
        except Exception:
            log.exception('system-image-cli exception')
            return 1
        # Say -c <no-such-channel> was given.  This will fail.
        if state.winner is None or len(state.winner) == 0:
            print('Already up-to-date')
        else:
            winning_path = [str(image.version) for image in state.winner]
            kws = dict(path=COLON.join(winning_path))
            if state.channel_switch is None:
                # We're not switching channels due to an alias change.
                template = 'Upgrade path is {path}'
            else:
                # This upgrade changes the channel that our alias is mapped
                # to, so include that information in the output.
                template = 'Upgrade path is {path} ({from} -> {to})'
                kws['from'], kws['to'] = state.channel_switch
            print(template.format(**kws))
        return
    else:
        # Run the state machine to conclusion.  Suppress all exceptions, but
        # note that the state machine will log them.  If an exception occurs,
        # exit with a non-zero status.
        log.info('running state machine [{}/{}]',
                 config.channel, config.device)
        try:
            list(state)
        except KeyboardInterrupt:
            return 0
        except Exception:
            log.exception('system-image-cli exception')
            return 1
        else:
            return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update channels."""

__all__ = [
    'Channels',
    ]


import json

from systemimage.helpers import Bag


def _parse_device_mappings(device_mapping):
    devices = {}
    # e.g. keys: nexus7, nexus4
    for device_name, mapping_1 in device_mapping.items():
        # Most of the keys at this level (e.g. index) have flat values,
        # however the keyring key is itself a mapping.
        keyring = mapping_1.pop('keyring', None)
        if keyring is not None:
            mapping_1['keyring'] = Bag(**keyring)
        # e.g. nexus7 -> {index, keyring}
        devices[device_name] = Bag(**mapping_1)
    return Bag(**devices)


class Channels(Bag):
    @classmethod
    def from_json(cls, data):
        mapping = json.loads(data)
        channels = {}
        for channel_name, mapping_1 in mapping.items():
            hidden = mapping_1.pop('hidden', None)
            if hidden is None:
                hidden = False
            else:
                assert hidden in (True, False), (
                    "Unexpected value for 'hidden': {}".format(hidden))
            mapping_1['hidden'] = hidden
            device_mapping = mapping_1.pop('devices')
            mapping_1['devices'] = _parse_device_mappings(device_mapping)
            channels[channel_name] = Bag(**mapping_1)
        return cls(**channels)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test persistent settings."""

__all__ = [
    'TestSettings',
    ]


import os
import unittest

from systemimage.config import Configuration
from systemimage.settings import Settings
from systemimage.testing.helpers import configuration


class TestSettings(unittest.TestCase):
    @configuration
    def test_creation(self, ini_file):
        config = Configuration()
        config.load(ini_file)
        self.assertFalse(os.path.exists(config.system.settings_db))
        settings = Settings()
        self.assertTrue(os.path.exists(config.system.settings_db))
        del settings
        # The file still exists.
        self.assertTrue(os.path.exists(config.system.settings_db))
        Settings()
        self.assertTrue(os.path.exists(config.system.settings_db))

    @configuration
    def test_get_set(self):
        settings = Settings()
        settings.set('permanent', 'waves')
        self.assertEqual(settings.get('permanent'), 'waves')

    @configuration
    def test_update(self):
        settings = Settings()
        settings.set('animal', 'ant')
        self.assertEqual(settings.get('animal'), 'ant')
        settings.set('animal', 'bee')
        self.assertEqual(settings.get('animal'), 'bee')

    @configuration
    def test_get_before_set(self):
        settings = Settings()
        self.assertEqual(settings.get('nothing'), '')

    @configuration
    def test_persistence(self):
        settings = Settings()
        settings.set('animal', 'cat')
        del settings
        self.assertEqual(Settings().get('animal'), 'cat')

    @configuration
    def test_prepopulated(self):
        # Some keys are pre-populated with default values.
        self.assertEqual(Settings().get('auto_download'), '1')



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""DBus API mediator."""


__all__ = [
    'Mediator',
    'Update',
    ]


import logging

from systemimage.helpers import last_update_date
from systemimage.state import State

log = logging.getLogger('systemimage')


class Update:
    """A representation of the available update."""

    def __init__(self, winners=None, error=''):
        self._winners = [] if winners is None else winners
        self.error = error

    @property
    def is_available(self):
        return len(self._winners) > 0

    @property
    def size(self):
        total_size = 0
        for image in self._winners:
            total_size += sum(filerec.size for filerec in image.files)
        return total_size

    @property
    def descriptions(self):
        return [image.descriptions for image in self._winners]

    @property
    def version(self):
        try:
            return str(self._winners[-1].version)
        except IndexError:
            # No winners.
            return ''

    @property
    def last_update_date(self):
        return last_update_date()


class Mediator:
    """This is the DBus API mediator.

    It essentially implements the entire DBus API, but at a level below the
    mechanics of DBus.  Methods of this class are hooked directly into the
    DBus layer to satisfy that interface.
    """

    def __init__(self, callback=None):
        self._state = State()
        self._update = None
        self._callback = callback

    def __repr__(self):
        return '<Mediator at 0x{:x} | State at 0x{:x}>'.format(
            id(self), id(self._state))

    def cancel(self):
        self._state.downloader.cancel()

    def pause(self):
        self._state.downloader.pause()

    def resume(self):
        self._state.downloader.resume()

    def check_for_update(self):
        """Is there an update available for this machine?

        :return: Flag indicating whether an update is available or not.
        :rtype: bool
        """
        if self._update is None:
            try:
                self._state.run_until('download_files')
            except Exception as error:
                # Rather than letting this percolate up, eventually reaching
                # the GLib main loop and thus triggering apport, Let's log the
                # error and set the relevant information in the class.
                log.exception('check_for_update failed')
                self._update = Update(error=str(error))
            else:
                self._update = Update(self._state.winner)
        return self._update

    def download(self):
        """Download the available update."""
        # We only want callback progress during the actual download.
        callback = self._state.downloader.callback
        try:
            self._state.downloader.callback = self._callback
            self._state.run_until('reboot')
        finally:
            self._state.downloader.callback = callback

    def reboot(self):
        """Issue the reboot."""
        # Transition through all remaining states.
        list(self._state)



# Copyright (C) 2013-2014 Canonical Ltd.
# Author: Barry Warsaw <barry@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Determine candidate images."""

__all__ = [
    'delta_filter',
    'full_filter',
    'get_candidates',
    'iter_path',
    ]


from collections import deque


class _Chaser:
    def __init__(self):
        # Paths are represented by lists, so we need to listify each element
        # of the initial set of roots.
        self._paths = deque()

    def __iter__(self):
        while self._paths:
            yield self._paths.pop()

    def push(self, new_path):
        # new_path must be a list.
        self._paths.appendleft(new_path)


def get_candidates(index, build):
    """Calculate all the candidate upgrade paths.

    This function returns a list of candidate upgrades paths, from the
    current build number to the latest build available in the index
    file.

    Each element of this list of candidates is itself a list of `Image`
    objects, in the order that they should be applied to upgrade the
    device.

    The upgrade candidate chains are not sorted, ordered, or prioritized
    in any way.  They are simply the list of upgrades that will satisfy
    the requirements.  It is possible that there are no upgrade candidates if
    the device is already at the latest build, or if the device is at a build
    too old to update.

    :param index: The index of available upgrades.
    :type index: An `Index`
    :param build: The build version number that the device is currently at.
    :type build: str
    :return: list-of-lists of upgrade paths.  The empty list is returned if
        there are no candidate paths.
    """
    # Start by splitting the images into fulls and delta.  Throw out any full
    # updates which have a minimum version greater than our version.
    fulls = set()
    deltas = set()
    for image in index.images:
        if image.type == 'full':
            if getattr(image, 'minversion', 0) <= build:
                fulls.add(image)
        elif image.type == 'delta':
            deltas.add(image)
        else:
            # BAW 2013-04-30: log and ignore.
            raise AssertionError('unknown image type: {}'.format(image.type))
    # Load up the roots of candidate upgrade paths.
    chaser = _Chaser()
    # Each full version that is newer than our current version provides the
    # start of an upgrade path.
    for image in fulls:
        if image.version > build:
            chaser.push([image])
    # Each delta with a base that matches our version also provides the start
    # of an upgrade path.
    for image in deltas:
        if image.base == build:
            chaser.push([image])
    # Chase the back pointers from the deltas until we run out of newer
    # versions.  It's possible to push new paths into the chaser if we find a
    # fork in the road (i.e. two deltas with the same base).
    paths = list()
    for path in chaser:
        current = path[-1]
        while True:
            # Find all the deltas that have this step as their base.
            next_steps = [delta for delta in deltas
                          if delta.base == current.version]
            # If there is no next step, then we're done with this path.
            if len(next_steps) == 0:
                paths.append(path)
                break
            # If there's only one next step, append that to path and keep
            # going, with this step as the current image.
            elif len(next_steps) == 1:
                current = next_steps[0]
                path.append(current)
            # Otherwise, we have a fork.  Take one fork now and push the other
            # paths onto the chaser.
            else:
                current = next_steps.pop()
                for fork in next_steps:
                    new_path = path.copy()
                    new_path.append(fork)
                    chaser.push(new_path)
                path.append(current)
    return paths


def iter_path(winner):
    """Iterate over all the file records for a given upgrade path.

    Image traversal will stop after the first `bootme` flag is seen, so the
    list of files to download may not include all the files in the upgrade
    candidate.

    :param winner: The list of images for the winning candidate.
    :return: A sequence of 2-tuples, where the first item is a "image
        number", i.e. which image in the path of winning images this file
        record belongs to, and the second item is the file record.
    """
    for n, image in enumerate(winner):
        for filerec in image.files:
            yield (n, filerec)
        if getattr(image, 'bootme', False):
            break


def full_filter(candidates):
    filtered = []
    for path in candidates:
        full_image = None
        for image in path:
            # Take the last full update we find from the start of the path.
            if image.type != 'full':
                break
            full_image = image
        if full_image is not None:
            filtered.append([full_image])
    return filtered


def delta_filter(candidates):
    filtered = []
    for path in candidates:
        new_path = []
        for image in path:
            # Add all the deltas from the start of the path to the first full.
            if image.type != 'delta':
                break
            new_path.append(image)
        if len(new_path) != 0:
            filtered.append(new_path)
    return filtered
